import math
import ROOT
import numpy as np
from raser.geometry import R2dDetector
from raser.geometry import R3dDetector

class MIPs:

    # MIPs particle
    def __init__(self):

        track_entry = [25.0,0]
        track_exit = [25.0,50.0]
        n_div = 100

        self.track_position = [ [] for n in range(n_div-1) ]
        self.delta_track = []

        # start position
        track_position_x=track_entry[0]
        track_position_y=track_entry[1]

        for i in range(n_div-1):

            x_div_point = (track_exit[0]-track_entry[0])/n_div*i+track_entry[0]+(track_exit[0]-track_entry[0])/(2*n_div)
            y_div_point = (track_exit[1]-track_entry[1])/n_div*i+track_entry[1]+(track_exit[1]-track_entry[1])/(2*n_div)

            self.track_position[i].append(x_div_point)
            self.track_position[i].append(y_div_point)

            self.delta_track.append(math.sqrt(math.pow(x_div_point-track_position_x,2)+math.pow(y_div_point-track_position_y,2)))

            track_position_x = x_div_point
            track_position_y = y_div_point
        
        self.mips_ionized()
    
    def mips_ionized(self):

        self.ionized_pairs = []
        self.ionized_total_pairs = 0.

        #Drift distance: self.delta_track
        loss_energy = 8.4 #silicon carbide /um
        Myf = ROOT.TFile("SiC_5um.root")
        hist = Myf.Get("Edep_device")

        for i in range(len(self.track_position)):
            
            gRandom = ROOT.TRandom3(0)
            ran_energy = hist.GetRandom(gRandom)
            n_pairs = self.delta_track[i]*ran_energy*1e6/(5*loss_energy)
            self.ionized_pairs.append(n_pairs)
            self.ionized_total_pairs += n_pairs


    def mips_landau(self):

        loss_energy = 8.4
        d_x = math.pow(self.track_exit[0]-self.track_entry[0],2)
        d_y = math.pow(self.track_exit[1]-self.track_entry[1],2)
        d_dis = math.sqrt(d_x+d_y)
        d_mpv = 0.04 * math.log(d_dis) + 0.27
        d_FWHM = 0.31 * math.pow(d_dis, -0.17)
        d_da = d_FWHM / 4.
        gRandom = ROOT.TRandom3(0)
        LanConst = gRandom.Landau(d_mpv, d_da)
        if (LanConst > 5.0 * d_mpv):
            LanConst = gRandom.Landau(d_mpv, d_da)
        Lan_pairs = LanConst*1000/loss_energy*d_dis

        return Lan_pairs

class TCTTracks():
    """
    Description:
        Transfer Carrier Distribution from Laser Coordinate System 
        to 2d Detector Coordinate System
    Parameters:
    ---------
    my_d : R2dDetector or R3dDetector
        the Detector
    laser : dict
        the Parameter List of Your Laser
    x_rel,y_rel,z_rel:
        the Normalized Coordinate for Laser Focus 
        in Detector Coordinate System
    @Modify:
    ---------
        2021/09/13
    """
    def __init__(self,my_d,laser,x_rel,y_rel,z_rel,min_carrier=0):
        self.tech=laser["tech"]
        self.direction=laser["direction"]

        self.refractionIndex=laser["refractionIndex"]

        self.wavelength=laser["wavelength"]*1e-3 #um
        self.tau=laser["tau"]
        self.power=laser["power"]
        self.widthBeamWaist=laser["widthBeamWaist"]#um
        
        self.r_step=laser["r_step"]#um
        self.h_step=laser["h_step"]#um

        if "l_Reyleigh" not in laser:
            self.l_Rayleigh = np.pi*self.widthBeamWaist**2*self.refractionIndex/self.wavelength
        else:
            self.l_Rayleigh = laser["l_Rayleigh"]#um

        if self.direction in ("top","bottom"):
            self.y_step=self.h_step
            self.z_step=self.x_step=self.r_step
            if self.direction == "top":
                def _getCarrierDensity(x,y,z):
                    return self.getCarrierDensity(y,x**2+z**2)
            if self.direction == "bottom":
                def _getCarrierDensity(x,y,z):
                    return self.getCarrierDensity(self.ly-y,x**2+z**2)

        elif self.direction == "edge":
            self.x_step=self.h_step
            self.z_step=self.y_step=self.r_step
            def _getCarrierDensity(x,y,z):
                return self.getCarrierDensity(x,y**2+z**2)
        else:
            raise NameError(self.direction)

        if isinstance(my_d,R2dDetector):
            self.lx=my_d.det_width#um
            self.ly=my_d.det_thin
            self.lz=my_d.det_width
        elif isinstance(my_d,R3dDetector):
            self.lx=my_d.l_x#um
            self.ly=my_d.l_y
            self.lz=my_d.l_z
        else:
            raise TypeError(my_d)

        if self.tech == "SPA":
            self.alpha=laser["alpha"]
            
        elif self.tech == "TPA":
            self.beta_2=laser["beta_2"]

        xArray = np.linspace(0.5*self.x_step,self.lx-0.5*self.x_step,int(self.lx/self.x_step))-self.lx*x_rel
        yArray = np.linspace(0.5*self.y_step,self.ly-0.5*self.y_step,int(self.ly/self.y_step))-self.ly*y_rel
        zArray = np.linspace(0.5*self.z_step,self.lz-0.5*self.z_step,int(self.lz/self.z_step))-self.lz*z_rel

        Y,X,Z=np.meshgrid(yArray,xArray,zArray) #Feature of numpy.meshgrid
        self.projGrid=_getCarrierDensity(X,Y,Z)\
            *self.x_step*self.y_step*self.z_step*1e-18

        if isinstance(my_d,R2dDetector):
            Y2d,X2d=np.meshgrid(yArray+self.ly*y_rel,xArray+self.lx*x_rel)
            self.projGrid=np.sum(self.projGrid,axis=2)
            self.track_position = list(np.transpose(np.array([list(np.ravel(X2d)),list(np.ravel(Y2d))])))
            self.ionized_pairs = list(np.ravel(self.projGrid))
            self.ionized_total_pairs = 0
            self.x_min,self.y_min = self.lx,self.ly
            self.x_max,self.y_max = 0,0
            for i in range(len(self.ionized_pairs)-1,-1,-1):
                if self.ionized_pairs[i]<=min_carrier:
                    del self.ionized_pairs[i]
                    del self.track_position[i]
                else:
                    self.ionized_total_pairs+=self.ionized_pairs[i]
                    self.x_min=min(self.x_min,self.track_position[i][0])
                    self.y_min=min(self.y_min,self.track_position[i][1])
                    self.x_max=max(self.x_max,self.track_position[i][0])
                    self.y_max=max(self.y_max,self.track_position[i][1])

        elif isinstance(my_d,R3dDetector):
            pass

        self.draw_nocarrier2D(x_rel,y_rel,z_rel,min_carrier)

    def getCarrierDensity(self,h,r2):
        widthSquared=(self.widthBeamWaist**2)*(1+(h/self.l_Rayleigh)**2)

        if self.tech=="SPA":
            intensity = ((2*self.power)/(np.pi*widthSquared*1e-12))*np.exp((-2*r2/(widthSquared)))*np.exp(-self.alpha*h*1e-6)
            e0 = 1.60217733e-19
            return self.alpha*intensity/(3.6*e0)
            
        elif self.tech=="TPA":
            k=(self.power**2)*8*np.log(2)/(self.tau*(np.pi**2.5)*(np.log(4))**0.5)
            intensity_squared = k*np.exp(-4*r2/widthSquared)/((widthSquared**2)*1e-24)
            h_Planck = 6.626*1e-34
            speedofLight = 2.998*1e8
            return self.beta_2*self.wavelength*1e-6*intensity_squared/(2*h_Planck*speedofLight)

    def draw_nocarrier2D(self,x_rel,y_rel,z_rel,min_carrier):
        c1 = ROOT.TCanvas("c1","canvas2",200,10,1000,1000)
        h = ROOT.TH2D("h","pairs of carrier generation",\
            int((self.x_max-self.x_min)/self.x_step)+1,self.x_min-0.5*self.x_step,self.x_max+0.5*self.x_step,\
            int((self.y_max-self.y_min)/self.y_step)+1,self.y_min-0.5*self.y_step,self.y_max+0.5*self.y_step)
        for i in range(len(self.track_position)):
            h.Fill(self.track_position[i][0], self.track_position[i][1],self.ionized_pairs[i])
        h.Draw()
        h.GetXaxis().SetTitle("Width [μm]")
        h.GetYaxis().SetTitle("Depth [μm]")
        c1.SaveAs("./fig/nocarrier_"\
            +str(round(x_rel,5))+"_"\
            +str(round(y_rel,5))+"_"\
            +str(round(z_rel,5))+"_"\
            +str(min_carrier)+".pdf")  



# -*- encoding: utf-8 -*-
'''
@Description: Calculate the weighting potential and electric field wiht fenics      
@Date       : 2021/08/31 15:04:25
@Author     : tanyuhang
@version    : 1.0
'''

import fenics
import mshr
import sys
import numpy as np
import ROOT
from array import array


#Calculate the weighting potential and electric field
class FenicsCal:

    def __init__(self,my_d,fen_dic):
        self.p_electric = []
        self.w_p_electric = []
        self.det_model = fen_dic['name']
        self.fl_x=my_d.l_x/fen_dic['xyscale']  
        self.fl_y=my_d.l_y/fen_dic['xyscale']  
        self.fl_z=my_d.l_z
        self.tol = 1e-14
        m_sensor_box=self.fenics_space(my_d)  
        self.mesh3D = mshr.generate_mesh(m_sensor_box,fen_dic['mesh'])  
        self.V = fenics.FunctionSpace(self.mesh3D, 'P', 1)
        self.fenics_p_electric(my_d)
        self.fenics_p_w_electric(my_d)

    def fenics_space(self,my_d):
        """
        @description: 
            Define the fenics solver space 
        @param:
            None
        @Returns:
            Fenics Box structure
        @Modify:
            2021/08/31
        """
        if "plugin3D" in self.det_model:
            self.sensor_range_confirm(my_d)
            m_sensor =  mshr.Box(fenics.Point(self.sx_l,self.sy_l, 0), 
                                 fenics.Point(self.sx_r,self.sy_r, self.fl_z))
            for i in range(len(my_d.e_tr)):
                e_t_i = my_d.e_tr[i]
                elec_n=mshr.Cylinder(fenics.Point(e_t_i[0],e_t_i[1],e_t_i[3]), 
                                     fenics.Point(e_t_i[0],e_t_i[1],e_t_i[4]),
                                     e_t_i[2],e_t_i[2])
                m_sensor =m_sensor - elec_n 
        elif "planar3D" in self.det_model:
            m_sensor =  mshr.Box(fenics.Point(0, 0, 0), 
                                 fenics.Point(self.fl_x, self.fl_y, self.fl_z))
        else:
            print("sensor model is wrrong")
            sys.exit()
        return m_sensor            
        
    def sensor_range_confirm(self,my_d):
        """
        @description:
            confirm the sensor range 
            at x,y axias to avoide the the big sensor size
            which will lead to the mesh complicate         
        @param:
            xv_min - fenics sensor x left value
            xv_max - fenics sensor x right value
            yv_min - fenics sensor y left value
            yv_max - fenics sensor y right value          
        @Modify:
            2021/08/31
        """    
        xv_list=[]
        yv_list=[]
        rest_length=50 #um
        length=0
        for i in range(len(my_d.e_tr)):
            e_t_i = my_d.e_tr[i]
            xv_list.append(e_t_i[0])
            yv_list.append(e_t_i[1])
            ele_radius= e_t_i[2]
        while length == 0:
            xv_max = max(xv_list)+ele_radius+rest_length 
            xv_min = min(xv_list)-ele_radius-rest_length 
            yv_max = max(yv_list)+ele_radius+rest_length
            yv_min = min(yv_list)-ele_radius-rest_length 
            if xv_max >= yv_max:
                yv_max = xv_max
            else:
                xv_max = yv_max
            if xv_min <= yv_min:
                yv_min = xv_min
            else:
                xv_min = yv_min
            if (xv_max > my_d.l_x or xv_min <0 
               or yv_max > my_d.l_y or yv_min < 0):
                rest_length -= 1
            else:
                length=1
        self.sx_l=xv_min 
        self.sx_r=xv_max 
        self.sy_l=yv_min 
        self.sy_r=yv_max 

    def fenics_p_electric(self,my_d):    
        """
        @description:
            Solve poisson equation to get potential and electric field
        @Modify:
            2021/08/31
        """
        if  "plugin3D" in self.det_model:
            bc_l=[]
            bc_l = self.boundary_definition_3D(my_d,"Possion")          
        elif "planar3D" in self.det_model:
            bc_l = self.boundary_definition_2D(my_d,"Possion")

        u = fenics.TrialFunction(self.V)
        v = fenics.TestFunction(self.V)
        f = fenics.Constant(self.f_value(my_d))
        a = fenics.dot(fenics.grad(u), fenics.grad(v))*fenics.dx
        L = f*v*fenics.dx
        # Compute solution
        self.u = fenics.Function(self.V)
        fenics.solve(a == L, self.u, bc_l,
                     solver_parameters=dict(linear_solver='gmres',
                                            preconditioner='ilu'))
        #calculate electric field
        W = fenics.VectorFunctionSpace(self.mesh3D, 'P', 1)
        self.E_field = fenics.project(fenics.as_vector((self.u.dx(0),
                                                        self.u.dx(1),
                                                        self.u.dx(2))),W)

    def fenics_p_w_electric(self,my_d):  
        """
        @description:
            Solve Laplace equation to 
            get weighting potential and weighting electric field
        @Modify:
            2021/08/31
        """
        if  "plugin3D" in self.det_model:
            bc_l = []
            bc_l = self.boundary_definition_3D(my_d,"Laplace")
        elif "planar3D" in self.det_model:
            bc_l = self.boundary_definition_2D(my_d,"Laplace")
        # Define variational problem
        u_w = fenics.TrialFunction(self.V)
        v_w = fenics.TestFunction(self.V)
        f_w = fenics.Constant(0)
        a_w = fenics.dot(fenics.grad(u_w), fenics.grad(v_w))*fenics.dx
        L_w = f_w*v_w*fenics.dx
        # Compute solution
        self.u_w = fenics.Function(self.V)
        fenics.solve(a_w == L_w, self.u_w, bc_l)

    def boundary_definition_3D(self,my_d,model):
        """
        @description:
            Get boundary definition of 3D detector with Possion and Laplace
        @Modify:
            2021/08/31
        """
        bc_l = []
        p_ele,n_ele=self.model_para(my_d,model)
        for i in range (len(my_d.e_tr)):
            e_i = my_d.e_tr[i]
            str_e = "x[0]>={e_0}-{e_2} && x[0]<={e_0}+"\
                    +"{e_2} && x[1]>={e_1}-{e_2} && "\
                    +"x[1]<={e_1}+{e_2} && x[2]>={e_3} \
                    && x[2]<={e_4} && on_boundary"
            elec_p = str_e.format(e_0=e_i[0], e_1=e_i[1],
                                  e_2=e_i[2], e_3=e_i[3],
                                  e_4=e_i[4])
            if e_i[5] == "p":
                bc = fenics.DirichletBC(self.V, p_ele, elec_p)
            else:
                bc = fenics.DirichletBC(self.V, n_ele, elec_p)
            bc_l.append(bc)
        return bc_l

    def boundary_definition_2D(self,my_d,model):
        """
        @description:
            Get boundary definition of 2D detector with Possion and Laplace
        @Modify:
            2021/08/31
        """
        p_ele,n_ele=self.model_para(my_d,model)
        u_D = fenics.Expression('x[2]<tol ? p_1:p_2',
                                degree=2, tol=1E-14,
                                p_1=p_ele, p_2=n_ele)
        def boundary(x, on_boundary):
            return abs(x[2])<self.tol or abs(x[2]-self.fl_z)<self.tol
        bc_l = fenics.DirichletBC(self.V, u_D, boundary)
        return bc_l
    
    def model_para(self,my_d,model):
        """
        @description:
            Choose Possion and Laplace parameters
        @Modify:
            2021/08/31
        """       
        bc_l=[]
        if model == "Possion":
            p_ele = my_d.v_voltage
            n_ele = 0.0
        elif model == "Laplace":
            p_ele = 0.0
            n_ele = 1.0
        else:
            print("The input fenics solver model is wrong")
        return p_ele,n_ele

    def f_value(self,my_d):
        """
        @description: 
            Cal f_value of Poisson equation
        @param:
            perm_mat -- Dielectric constant of using material
                     -- 11.7 Silicon
                     -- 9.76 Silicon carbide
        @Modify:
            2021/08/31
        """
        if my_d.mater == 0:
            perm_mat = 11.7  
        elif my_d.mater == 1:
            perm_mat = 9.76  
        else:
            print("material is wrong")            
        e0 = 1.60217733e-19
        perm0 = 8.854187817e-12   #F/m
        f_value = e0*my_d.d_neff*1e6/perm0/perm_mat
        return f_value
        
    def get_e_field(self,px,py,pz):
        """
        @description: 
            Get eletric field at the px,py,pz position
        @param:
            out_range -- out_range = False
                      -- Position (x,y,z) don't exit in sensor fenics range
        @reture:
            Eletric field along x,y,z direction
        @Modify:
            2021/08/31
        """
        out_range=self.judge_fenics_range(px,py,pz)
        if out_range:   
            x_value,y_value,z_value = 0,0,0
        else:
            scale_px=px%self.fl_x
            scale_py=py%self.fl_y          
            try:
                x_value,y_value,z_value = self.E_field(scale_px,scale_py,pz)
            except RuntimeError:
                x_value,y_value,z_value = 0,0,0
        return x_value,y_value,z_value

    def get_w_p(self,px,py,pz):
        """
        @description: 
            Get weighting potential at the (x,y,z) position
        @param:
            out_range -- out_range = False
                      -- Position (x,y,z) don't exit in sensor fenics range
        @reture:
            Get weighting potential at (x,y,z) position
        @Modify:
            2021/08/31
        """
        out_range=self.judge_fenics_range(px,py,pz)
        if out_range:   
            f_w_p = 1.0
        else:
            scale_px=px%self.fl_x
            scale_py=py%self.fl_y   
            try:
                f_w_p = self.u_w(scale_px,scale_py,pz)
            except RuntimeError:
                f_w_p = 0.0
        return f_w_p

    def get_potential(self,px,py,pz):
        """
        @description: 
            Get potential at the (x,y,z) position
        @param:
            out_range -- out_range = False
                      -- Position (x,y,z) don't exit in sensor fenics range
        @reture:
            Get potential at (x,y,z) position
        @Modify:
            2021/08/31
        """
        out_range=self.judge_fenics_range(px,py,pz)
        if out_range:   #
            f_w_p = 0
        else:
            scale_px=px%self.fl_x
            scale_py=py%self.fl_y   
            try:
                f_w_p = self.u(scale_px,scale_py,pz)
            except RuntimeError:
                f_w_p = 0.0
        return f_w_p

    def judge_fenics_range(self,px,py,pz):
        """
        @description: 
           Judge whether (x,y,z) position is in sensor fenics range
        @reture:
            False: not
            Ture:  in
        @Modify:
            2021/08/31
        """
        if "plugin3D" in self.det_model:
            if (px < self.sx_l or px > self.sx_r 
                   or py < self.sy_l or py > self.sy_r):
                out_range=True
            else:
                out_range=False
        elif "planar3D" in self.det_model:
            out_range=False
        return out_range
    
    def __del__(self):
        pass


class FenicsCal2D:

    def __init__(self,det):
        
        self.det = det

        # poential & field
        self.potential_value_2d = []
    
        self.electric_field_x_value = [ [] for n in range(self.det.ny+1) ]
        self.electric_field_y_value = [ [] for n in range(self.det.ny) ]

        self.electric_field_x_position = [ [] for n in range(self.det.ny+1) ]
        self.electric_field_y_position = [ [] for n in range(self.det.ny) ]

        # weighting poential & weighting field
        self.weighting_potential_value_2d = []
    
        self.weighting_electric_field_x_value = [ [] for n in range(self.det.ny+1) ]
        self.weighting_electric_field_y_value = [ [] for n in range(self.det.ny) ]

        self.weighting_electric_field_x_position = [ [] for n in range(self.det.ny+1) ]
        self.weighting_electric_field_y_position = [ [] for n in range(self.det.ny) ]

        self.solve()
        #self.draw()

    def cal_possion(self):
        
        width = self.det.det_width
        thin = self.det.det_thin
        
        nx = self.det.nx
        ny = self.det.ny

        # Create mesh and function space
        mesh = fenics.RectangleMesh(fenics.Point(0, 0), fenics.Point(width, thin), nx, ny)
        V = fenics.FunctionSpace(mesh, "P", 1)

        # Define boundary condition
        u_D = fenics.Expression('x[1] < tol? det_voltage : 0', degree = 2,tol = 1E-14,det_voltage = self.det.bias_voltage)

        def boundary(x, on_boundary):
            return abs(x[1])<1E-14 or abs(x[1]-thin)<1E-14

        bc = fenics.DirichletBC(V, u_D, boundary)

        # Define variational problem
        u = fenics.TrialFunction(V)
        v = fenics.TestFunction(V)

        f = fenics.Expression(self.det.doping_epr,degree=2)
        a = fenics.dot(fenics.grad(u), fenics.grad(v))*fenics.dx
        L = f*v*fenics.dx #+ g*v*ds

        # Compute solution
        u = fenics.Function(V)
        fenics.solve(a == L, u, bc)

        potential_value_1d = u.compute_vertex_values()
        potential_value_2d = np.array(potential_value_1d).reshape(ny+1,nx+1)

        self.potential_value_2d = potential_value_2d

        # print(self.potential_value_2d)


    def cal_weighting_possion(self):

        width = self.det.det_width
        thin = self.det.det_thin
        
        nx = self.det.nx
        ny = self.det.ny

        # Create mesh and function space
        mesh = fenics.RectangleMesh(fenics.Point(0, 0), fenics.Point(width, thin), nx, ny)
        V = fenics.FunctionSpace(mesh, "P", 1)

        # Define boundary condition
        u_D = fenics.Expression('x[1] < tol? det_voltage : 0', degree = 2,tol = 1E-14,det_voltage = self.det.bias_voltage/abs(self.det.bias_voltage))

        def boundary(x, on_boundary):
            return abs(x[1])<1E-14 or abs(x[1]-thin)<1E-14

        bc = fenics.DirichletBC(V, u_D, boundary)

        # Define variational problem
        u = fenics.TrialFunction(V)
        v = fenics.TestFunction(V)

        f = fenics.Constant(0)
        a = fenics.dot(fenics.grad(u), fenics.grad(v))*fenics.dx
        L = f*v*fenics.dx #+ g*v*ds

        # Compute solution
        u = fenics.Function(V)
        fenics.solve(a == L, u, bc)

        weighting_potential_value_1d = u.compute_vertex_values()
        weighting_potential_value_2d = np.array(weighting_potential_value_1d).reshape(ny+1,nx+1)

        self.weighting_potential_value_2d = weighting_potential_value_2d

        return weighting_potential_value_2d

    def cal_electric_field(self):

        width = self.det.det_width
        thin = self.det.det_thin
        
        nx = self.det.nx
        ny = self.det.ny

        x_step = width/nx
        y_step = thin/ny

        # x direction
        for j in range(ny+1):
            for i in range(nx):

                # electric field
                tmp_xpos = 0.5*x_step*(2*i+1)
                tmp_xef = (self.potential_value_2d[j][i] - self.potential_value_2d[j][i+1])/x_step
                self.electric_field_x_position[j].append(tmp_xpos)
                self.electric_field_x_value[j].append(tmp_xef)

                # weighting field
                tmp_wxpos = 0.5*x_step*(2*i+1)
                tmp_xwef = (self.weighting_potential_value_2d[j][i] - self.weighting_potential_value_2d[j][i+1])/x_step
                self.weighting_electric_field_x_position[j].append(tmp_wxpos)
                self.weighting_electric_field_x_value[j].append(tmp_xwef)


        # y direction 
        for j in range(ny):
            for i in range(nx+1):

                # electric field
                tmp_ypos = 0.5*y_step*(2*j+1)
                tmp_yef = (self.potential_value_2d[j][i]- self.potential_value_2d[j+1][i])/y_step
                self.electric_field_y_position[j].append(tmp_ypos)
                self.electric_field_y_value[j].append(tmp_yef)

                # weighting field
                tmp_wypos = 0.5*y_step*(2*j+1)
                tmp_ywef = (self.weighting_potential_value_2d[j][i] - self.weighting_potential_value_2d[j+1][i])/y_step
                self.weighting_electric_field_y_position[j].append(tmp_wypos)
                self.weighting_electric_field_y_value[j].append(tmp_ywef)     



    def cal_point_field(self,px_point,py_point,input_value):

        width = self.det.det_width
        thin = self.det.det_thin
        
        nx = self.det.nx
        ny = self.det.ny

        x_step = width/nx
        y_step = thin/ny

        #Interpolation method 
        rex_value=px_point%x_step
        nx_value=int(px_point/x_step)
        rey_value=py_point%y_step
        ny_value=int(py_point/y_step)

        if(rex_value>x_step/2):
            e_v_x1=rex_value-x_step/2
            nx1_v=nx_value
            nx2_v=nx_value+1
        else:
            e_v_x1=rex_value+x_step/2
            e_v_x2=x_step-e_v_x1
            nx1_v=nx_value-1
            nx2_v=nx_value

        if(rey_value>y_step/2):
            e_v_y1=rey_value-y_step/2
            ny1_v=ny_value
            ny2_v=ny_value+1
        else:
            e_v_y1=rey_value+y_step/2
            e_v_y2=y_step-e_v_y1
            ny1_v=ny_value-1
            ny2_v=ny_value

        if (nx_value<=0):
            r_u=0
            nx1_v=nx2_v
        elif (nx_value>=nx-1):
            r_u=0
            nx2_v=nx1_v
        else:
            r_u=e_v_x1/x_step

        if (ny_value<=0):
            r_t=0
            ny1_v=ny2_v
        elif (ny_value>=ny-1):
            r_t=0
            ny2_v=ny1_v
        else:
            r_t=e_v_y1/y_step

        value_11=input_value[ny1_v][nx1_v]
        value_21=input_value[ny1_v][nx2_v]
        value_12=input_value[ny1_v][nx1_v]
        value_22=input_value[ny1_v][nx2_v]
        out_field=0.0
        out_field=(1-r_u)*(1-r_t)*value_11
        out_field+=r_u*(1-r_t)*value_21
        out_field+=r_u*r_t*value_22
        out_field+=(1-r_u)*r_t*value_12

        return out_field        


    def solve(self):

        self.cal_possion()
        self.cal_weighting_possion()
        self.cal_electric_field()

    def draw(self):

        cutline = int(self.det.nx/2.0)

        # plot electric field at x = middle
        ep = array( 'd' )
        ev = array( 'd' )

        for i in range(self.det.ny):

            ep.append(self.electric_field_y_position[i][cutline])

            efx = self.cal_point_field(self.det.det_width/2.0, self.electric_field_y_position[i][cutline], self.electric_field_x_value)
            efy = self.cal_point_field(self.det.det_width/2.0, self.electric_field_y_position[i][cutline], self.electric_field_y_value)
            ef = np.array([efx,efy])

            ef_value = np.linalg.norm(ef)#*1e4 #V/cm

            ev.append(ef_value)

        # print(ep)
        # print(ev)

        g_e = ROOT.TGraph(self.det.ny,ep,ev)

        g_e.SetLineColor(600)
        g_e.SetLineWidth(4)
        g_e.SetTitle( 'Electric Field at Cut Line' )
        g_e.GetXaxis().SetTitle( 'Depth [um]' )
        g_e.GetXaxis().SetRangeUser(0,self.det.det_thin)
        g_e.GetYaxis().SetTitle( 'E [V/um]' )

        g_e.GetXaxis().CenterTitle()
        g_e.GetXaxis().SetTitleOffset(1.8)
        g_e.GetXaxis().SetTitleSize(0.05)
        g_e.GetXaxis().SetLabelSize(0.05)
        #g_e.GetXaxis().SetNdivisions(505)

        g_e.GetYaxis().CenterTitle()
        g_e.GetYaxis().SetTitleOffset(1.8)
        g_e.GetYaxis().SetTitleSize(0.05)
        g_e.GetYaxis().SetLabelSize(0.05)
        #g_e.GetYaxis().SetNdivisions(505)
       
        c = ROOT.TCanvas( 'c', 'c',500, 500 )
        c.SetGrid()
        c.SetLeftMargin(0.18)
        c.SetBottomMargin(0.18)
        # c.Divide(1,2)

        c.cd()
        g_e.Draw()
        c.Modified()
        c.Update()
        c.SaveAs("./fig/%s_%s_electricfield_%dV.pdf"%(self.det.mat_name,self.det.det_model,self.det.bias_voltage))



# -*- encoding: utf-8 -*-
'''
@Description: Detector structure definition      
@Date       : 2021/08/31 11:09:40
@Author     : tanyuhang
@version    : 1.0
''' 

import ROOT
import math
import sys
from raser.model import Material

#Detector structure
class R3dDetector:
    def __init__(self,dset):
        """
        Description:
            Differnet types detectors parameters assignment.
        Parameters:
        ---------
        det_dic : dictionary
            Contain all detector parameters 
        meter : int
            mater = 1, SiC
            mater = 0, Si          
        Modify:
        ---------
            2021/09/02
        """ 
        det_dic = dset.detector
        self.l_x = det_dic['lx'] 
        self.l_y = det_dic['ly']  
        self.l_z = det_dic['lz'] 
        self.d_neff = det_dic['doping'] 
        self.v_voltage = det_dic['voltage'] 
        self.temperature = det_dic['temp']
        self.det_model = dset.det_model
        self.mater = 1    
        self.current_define()
        if 'plugin3D' in self.det_model:
            self.set_3D_electrode(det_dic['e_ir'],det_dic['e_gap'])

    def current_define(self):
        """
        @description: 
            Parameter current setting     
        @param:
            matter -- 0 is silicon, 1 is silicon carbide
            positive_cu -- Current from holes move
            negative_cu -- Current from electrons move
            sum_cu -- Current from e-h move
        @Returns:
            None
        @Modify:
            2021/08/31
        """
        self.t_bin = 50e-12
        self.t_end = 6.0e-9
        self.t_start = 0
        self.n_bin = int((self.t_end-self.t_start)/self.t_bin)
        
        self.positive_cu = ROOT.TH1F("charge+", "Positive Current",
                                     self.n_bin, self.t_start, self.t_end)
        self.negative_cu = ROOT.TH1F("charge-", "Negative Current",
                                     self.n_bin, self.t_start, self.t_end)
        self.sum_cu = ROOT.TH1F("charge","Total Current",
                                self.n_bin, self.t_start, self.t_end)

    def set_3D_electrode(self,e_ir,e_gap=0):
        """
        @description: 
            3D plug-in detector electrodes setting     
        @param:
            e_ir -- The radius of electrode
            e_gap -- The spacing between the electrodes  
        @Returns:
            None
        @Modify:
            2021/08/31
        """
        self.e_gap = e_gap
        e_r = e_ir  
        e_int = e_gap 
        e_t_y = self.infor_ele(e_r,e_int)
        self.e_tr=[]
        self.e_t_1 = [self.l_x*0.5          ,self.l_y*0.5      ,e_r,0,self.l_z,"n"]
        self.e_t_2 = [self.l_x*0.5-e_int    ,self.l_y*0.5      ,e_r,0,self.l_z,"p"]
        self.e_t_3 = [self.l_x*0.5+e_int    ,self.l_y*0.5      ,e_r,0,self.l_z,"p"]
        self.e_t_4 = [self.l_x*0.5-e_int*0.5,self.l_y*0.5+e_t_y,e_r,0,self.l_z,"p"]
        self.e_t_5 = [self.l_x*0.5+e_int*0.5,self.l_y*0.5+e_t_y,e_r,0,self.l_z,"p"]
        self.e_t_6 = [self.l_x*0.5-e_int*0.5,self.l_y*0.5-e_t_y,e_r,0,self.l_z,"p"]
        self.e_t_7 = [self.l_x*0.5+e_int*0.5,self.l_y*0.5-e_t_y,e_r,0,self.l_z,"p"]
        for i in range(7):
           n_e = eval('self.e_t_' + str(i+1))
           self.e_tr.append(n_e)

    def infor_ele(self,e_r,e_int):
        """
        @description: 
            3D plug-in detector electrodes spacing    
        @param:
            e_x_gap -- Judge whether the electrode is outer the detector
            e_t_y -- Distance between electrodes at y bottom or to and center
        @Returns:
            None
        @Modify:
            2021/08/31
        """
        e_x_gap = self.l_x - 2*e_r - 2*e_int
        if e_x_gap < 0:
            print("the electrode at x position is large than sensor length")
            sys.exit(0)
        e_t_y = math.sqrt(e_int*e_int*0.75)
        if 2*e_t_y > self.l_y:
            print("the electrode at y position is large than sensor length")
            sys.exit(0)            
        return e_t_y



class R2dDetector:

    def __init__(self,det_dic):

        self.det_model = det_dic["name"]
        self.det_width = det_dic["det_width"] #um X-axis
        self.det_thin = det_dic["det_thin"] #um Y-axis
        self.x_step = det_dic["x_step"]
        self.y_step = det_dic["y_step"]
        self.mat = Material(det_dic["material"])
        self.doping_epr = det_dic["doping_epr"]
        self.bias_voltage = det_dic["bias_voltage"]
        self.temperature = det_dic["temperature"]

        self.mat_name = self.mat.mat_name
        self.par_dict = self.mat.mat_database()
        self.nx = int(self.det_width/self.x_step)
        self.ny = int(self.det_thin/self.y_step)

        self.t_start = 0
        self.t_end = 3e-9
        self.t_bin = 50e-12
        self.n_bin = int((self.t_end-self.t_start)/self.t_bin)
        self.positive_cu = ROOT.TH1F("charge+","Positive Current",self.n_bin,0,self.t_end)
        self.negative_cu = ROOT.TH1F("charge-","Negative Current",self.n_bin,0,self.t_end)

        self.gain_positive_cu = ROOT.TH1F("gain_charge+","Current Contribution",self.n_bin,0,self.t_end)
        self.gain_negative_cu = ROOT.TH1F("gain_charge-","Gain Negative Current",self.n_bin,0,self.t_end)

        self.gain_n_n_cu = ROOT.TH1F("gain_n_n","Current Contribution",self.n_bin,0,self.t_end)
        self.gain_n_p_cu = ROOT.TH1F("gain_n_p","Current Contribution",self.n_bin,0,self.t_end)
        self.gain_p_n_cu = ROOT.TH1F("gain_p_n","Current Contribution",self.n_bin,0,self.t_end)
        self.gain_p_p_cu = ROOT.TH1F("gain_p_p","Current Contribution",self.n_bin,0,self.t_end)

        self.sum_cu = ROOT.TH1F("charge","Total Current",self.n_bin,0,self.t_end)
    
    def mesh(self,x_step,y_step):

        self.x_step = x_step
        self.y_step = y_step

        self.nx = int(self.det_width/x_step)
        self.ny = int(self.det_thin/y_step)

    def set_mat(self,mat):
        self.mat =mat
        self.mat_name = mat.mat_name
        self.par_dict = self.mat.mat_database()

    def set_doping(self,doping_epr):
        self.doping_epr = doping_epr

    def set_bias_voltage(self,bias_voltage):
        self.bias_voltage = bias_voltage

    def set_temperature(self,temperature):
        self.temperature = temperature



# -*- encoding: utf-8 -*-

'''
Description: Raser parameter settings      
@Date       : 2021/09/02 09:57:00
@Author     : tanyuhang
@version    : 1.0
'''
import json

# Define all input parameters used in raser main process
class Setting:
    def __init__(self,parameters):
        """
        Description:
            1.Different functions detector(), fenics(), pygeant4() define 
            different class parameters.
            2.Parameter defined according input parameters.
        Parameters:
        ---------
        det_model : str
            Define the sensor models for simulation.eg. planar-3D plugin-3D
        _pardic : dictionaries
            Storage the input parameters
        steplength : float
            The length of  each step for e-h pairs drift
        laser_model : str
            Define the Laser Absorption Pattern 
        
        @Modify:
        ---------
            2021/09/02
        """
        self._pardic = {}
        self.input2dic(parameters)
        self.det_model = self._pardic['det_model']
        self.read_par(self._pardic['parfile'])
        if "laser_model" in self._pardic:
            self.laser_model=self._pardic['laser_model']
            self.append_par(self._pardic['laser_file'])
        if self.det_model in ("plugin3D","planar3D"):
            self.scan_variation()

    def input2dic(self,parameters):
        " Transfer input list to dictinary"
        for par in parameters:
            name,_,value=par.rpartition('=')
            self._pardic[name]=value

    def read_par(self,jsonfile):
        "Read the setting.json file and save the input parametersin paras"
        with open(jsonfile) as f:
            dic_pars = json.load(f)
        for dic_par in dic_pars:
            if dic_par['name'] in self.det_model:
                self.steplength = float(dic_par['steplength'])
                paras =  dic_par
        for x in paras: 
            if self.is_number(paras[x]):          
                paras[x] = float(paras[x])
            else:
                paras[x] = paras[x]
        self.paras = paras

    def append_par(self,jsonfile):
        "Read the laser.json file and save the input parameters in paras"
        with open(jsonfile) as f:
            dic_pars = json.load(f)
        for dic_par in dic_pars:
            for x in dic_par: 
                if self.is_number(dic_par[x]):          
                    dic_par[x] = float(dic_par[x])
                else:
                    dic_par[x] = dic_par[x]
            self.paras.update(dic_par) 

    @property
    def detector(self):
        """
        Description:
            Define differnet types detectors parameters. Like:
            planar3D, plugin3D
        Parameters:
        ---------
        lx,ly,lz : float
            Detector length, width and height
        doping : float
            Doping concentation should times 1e12 /um^3  
            -- N-type is positive (negetive volatge applied) 
            -- P-type is negetive (positive volatge applied)
        temp : float
            Tempareture
        e_ir : float
            Radius of electrode in 3D
        e_gap : float
            Spacing between the electrodes in 3D
        @Returns:
        ---------
            A dictionary containing all parameters used in detector  
        @Modify:
        ---------
            2021/09/02
        """
        p = self.paras
        if "planar3D" in self.det_model:
            detector = {'name':'planar3D', 'lx':p['lx'], 'ly':p['ly'], 
                        'lz':p['lz'], 'doping':p['doping'], 
                        'voltage':p['voltage'], 'temp':p['temp']
                        }
            
        if "plugin3D" in self.det_model:
            detector = {'name':'plugin3D', 'lx':p['lx'], 'ly':p['ly'], 
                        'lz':p['lz'], 'doping':p['doping'], 
                        'voltage':p['voltage'], 'temp':p['temp'], 
                        'e_ir':p['e_ir'], 'e_gap':p['e_gap']
                        }
        
        if "lgad2D" in self.det_model:
            detector = {'name':'lgad2D',
                        'det_width':p['det_width'], 'det_thin':p['det_thin'],
                        'x_step':p['x_step'], 'y_step':p['y_step'],
                        'material':p['material'],
                        'doping_epr':p['doping_epr'],
                        'bias_voltage':p['bias_voltage'],
                        'temperature':p['temperature']
                        }

        if "pin2D" in self.det_model:
            detector = {'name':'pin2D',
                        'det_width':p['det_width'], 'det_thin':p['det_thin'],
                        'x_step':p['x_step'], 'y_step':p['y_step'],
                        'material':p['material'],
                        'doping_epr':p['doping_epr'],
                        'bias_voltage':p['bias_voltage'],
                        'temperature':p['temperature']
                        }
        return detector

    @property
    def fenics(self):
        """
        Description:
            Define differnet fenics parameters
        Parameters:
        ---------
        mesh : int
            Mesh precision value, the bigger the higher the accuracy
        xyscale : int
            In plane detector, scale_xy is scaling sensor 50 times at x and 
            y axis, so the precision can improve 50 times in echo distance 

        @Returns:
        ---------
            A dictionary containing all parameters used in fenics  
        @Modify:
        ---------
            2021/09/02
        """
        p = self.paras
        if "planar3D" in self.det_model:
            fenics = {'name':'planar3D', 
                      'mesh':p['mesh'], "xyscale":p['xyscale']}
        if "plugin3D" in self.det_model:
            fenics = {'name':'plugin3D', 
                      'mesh':p['mesh'], "xyscale":p['xyscale']}
        return fenics

    @property
    def pygeant4(self):
        """
        Description:
            Define differnet geant4 parameters
        Parameters:
        ---------
        maxstep : float
            Simulate the step size of each step in Geant4
        par_in : list
            Incident particle position
        par_out : list
            Theoretical position of outgoing particles
        g4_vis : bool
            False: Graphical interface of geant4 particles Disabled
            True: Graphical interface of geant4 particles Enabled
        @Returns:
        ---------
            A dictionary containing all parameters used in geant4  
        @Modify:
        ---------
            2021/09/02
        """
        p = self.paras
        if "planar3D" in self.det_model:
            pygeant4 = {'name':'planar3D',
                        'maxstep':p['maxstep'], 'g4_vis':p['g4_vis'],
                        'par_in':[p['par_inx'], p['par_iny'], p['par_inz']], 
                        "par_out":[p['par_outx'], p['par_outy'], p['par_outz']],
                        }
        if "plugin3D" in self.det_model:
            pygeant4 = {'name':'plugin3D', 
                        'maxstep':p['maxstep'], 'g4_vis':p['g4_vis'],
                        'par_in':[p['par_inx'], p['par_iny'], p['par_inz']], 
                        "par_out":[p['par_outx'], p['par_outy'], p['par_outz']],
                        }
        return pygeant4

    # @property
    # def mips(self):
    #     """
    #     Description:
    #         Define mips parameters
    #     Parameters:
    #     ---------
    #     track_entry : list
    #         Incident particle position
    #     track_exit : list
    #         Exit position
    #     n_div: int
    #         Divide the track line to n_div points
    #     @Returns:
    #     ---------
    #         A dictionary containing all parameters
    #     @Modify:
    #     ---------
    #         2021/09/07
    #     """
    #     p = self.paras
    #     track_par = {'name':'mips',
    #                  'track_entry':[25,0],
    #                  'track_exit':[25,50],
    #                  'n_div':100}      
    #     return track_par

    @property
    def laser(self):
        """
        Description:
            Define laser parameters
        Parameters:
        ---------
        tech : str
            Interaction Pattern Between Laser and Detector
        direction : str
            Direction of Laser Incidence, Could be "top" "edge" or "bottom"

        alpha : float
            the Linear Absorption Coefficient of the Bulk of the Device
        beta_2 : float
            the Quadratic Absorption Coefficient of the Bulk of the Device
        refractionIndex :float
            the Refraction Index of the Bulk of the Device

        wavelength : float
            the Wavelength of Laser in nm
        tau : float
            the Full-width at Half-maximum (FWHM) of the Beam Temporal Profile
        power : float
            the Energy per Laser Pulse
        widthBeamWaist : float
            the Width of the Beam Waist of the Laser in um
        l_Rayleigh : float
            the Rayleigh Width of the Laser Beam

        r_step, h_step : float
            the Step Length of Block in um,
            Carriers Generated in the Same Block Have the Same Drift Locus
        @Returns:
        ---------
            A dictionary containing all parameters used in TCTTracks 
        @Modify:
        ---------
            2021/09/08
        """
        p = self.paras
        if hasattr(self,"laser_model"):
            laser = {'tech':p['laser_model'],'direction':p['direction'],
                    'refractionIndex':p['refractionIndex'],
                    "wavelength":p["wavelength"],"tau":p["tau"],"power":p["power"],"widthBeamWaist":p["widthBeamWaist"],
                    'r_step':p['r_step'],'h_step':p['h_step']
                    }
            if self.laser_model == "SPA":
                laser.update({'alpha':p['alpha']})
            if self.laser_model == "TPA":
                laser.update({'beta_2':p['beta_2']})
            if 'l_Rayleigh' in p:
                laser.update({'l_Rayleigh':p['l_Rayleigh']})
        return laser

    @property
    def amplifer(self):
        """
        Description:
            Define differnet amplifers parameters
        Parameters:
        ---------
        maxstep : float
            Simulate the step size of each step in Geant4
        par_in : list
            Incident particle position
        par_out : list
            Theoretical position of outgoing particles
        g4_vis : bool
            False: Do not display graphical interface of geant4 partivles
            True: Do not display graphical interface of geant4 partivles
        @Returns:
        ---------
            A dictionary containing all parameters used in geant4  
        @Modify:
        ---------
            2021/09/02
        """
        p = self.paras
        CSA_par = {'name':'CSA_ampl', 't_rise':p['t_rise'], 
                   't_fall':p['t_fall'], 'trans_imp':p['trans_imp'], 
                   'CDet':p['CDet']
                  }
        BB_par = {'name':'BB_ampl', 'BBW':p['BBW'], 
                  'BBGain':p['BBGain'], 'BB_imp':p['BB_imp'],'OscBW':p['OscBW']
                 }
        return [CSA_par,BB_par]

    def scan_variation(self):
        " Define parameters of batch mode"
        if "3Dscan" in self.det_model:
            self.total_events = int(self._pardic['total_e'])
            self.intance_number = int(self._pardic['instan'])
            self.g4seed = self.intance_number * self.total_events
            self.output = self._pardic["output"]
        else:
            p = self.paras
            self.total_events = int(p['total_events'])
            self.g4seed = 0 

    def is_number(self,s):
        "Define whethre the s is a number or not"
        try:
            float(s)
            return True
        except ValueError:
            pass
        try:
            import unicodedata
            unicodedata.numeric(s)
            return True
        except (TypeError, ValueError):
            pass
        return False 


# -*- encoding: utf-8 -*-
'''
Description:  Simulate the e-h pairs drift and calculate induced current     
@Date       : 2021/09/02 14:01:46
@Author     : tanyuhang
@version    : 1.0
'''

import random
import numpy as np
import ROOT
import math
import sys
from array import array
from raser.model import Mobility
from raser.model import Avalanche
from raser.source import MIPs

#The drift of generated particles
class CalCurrent:
    def __init__(self, my_d, my_f, my_g4p, dset, batch=0):
        #mobility related with the magnetic field (now silicon useless)
        self.muhh=1650.0   
        self.muhe=310.0
        self.BB=np.array([0,0,0])
        self.sstep=dset.steplength #drift step
        print(self.sstep)
        self.kboltz=8.617385e-5 #eV/K
        self.max_drift_len=1e9 #maximum diftlenght [um]
        self.parameters(my_g4p, my_d, batch)
        self.ionized_drift(my_g4p,my_f,my_d)
            
    def parameters(self,my_g4p, my_d, batch): 
        """" Define the output dictionary """   
        self.d_dic_n = {}
        self.d_dic_p = {}
        self.events_position(my_g4p, batch) 
        for n in range(len(self.tracks_p)-1):
            self.d_dic_n["tk_"+str(n+1)] = [ [] for n in range(5) ]
            self.d_dic_p["tk_"+str(n+1)] = [ [] for n in range(5) ]
           
    def events_position(self,my_g4p, batch):
        """
        Description:
            Events position and energy depositon
        Parameters:
        ---------
        batch : int
            batch = 0: Single event, choose hit electron
            batch != 0: Multi event, batch is the number of electron     
        @Modify:
        ---------
            2021/09/13
        """     
        if batch == 0:
            for j in range(len(my_g4p.p_steps)):
                if len(my_g4p.p_steps[j])>10 and batch == 0:
                    self.beam_number = j
                    self.tracks_p = my_g4p.p_steps[j]
                    self.tracks_step_edep = my_g4p.energy_steps[j]
                    self.tracks_t_edep = my_g4p.edep_devices[j]
                    batch+=1
                    break
            if batch == 0:
                print("the sensor did have the hit particles")
                sys.exit()
        else:
            self.beam_number = batch
            self.tracks_p = my_g4p.p_steps[batch]
            self.tracks_step_edep = my_g4p.energy_steps[batch]
            self.tracks_t_edep = my_g4p.edep_devices[batch]

    def ionized_drift(self,my_g4p,my_f,my_d):
        """
        Description:
            The drift simulation of all tracks
            One track is corresponding to one electron-hole pair 
        """
        for i in range(len(self.tracks_p)-1):
            self.n_track=i+1
            for j in range(2):
                if (j==0):
                    self.charg=1 #hole
                if (j==1):
                    self.charg=-1 #electron 
                self.loop_electon_hole(my_g4p,my_f,my_d,i)
        self.get_current(my_d)

    def energy_deposition(self,my_d,j):
        """" Deposition energy and generate e-h pair """
        sic_loss_e=self.meter_choose(my_d)
        n_pairs=self.tracks_step_edep[j]*1e6/sic_loss_e
        return n_pairs 

    def loop_electon_hole(self,my_g4p,my_f,my_d,i):
        """
        Description:
            Loop and record the induced cuurent of each eletron or holes         
        Parameters:
        ---------
        arg1 : int
            
        @Modify:
        ---------
            2021/09/13
        """
        self.initial_parameter()
        self.d_x=self.tracks_p[i+1][0]
        self.d_y=self.tracks_p[i+1][1]
        self.d_z=self.tracks_p[i+1][2] - my_g4p.init_tz_device 
        while (self.end_cond == 0):
            if self.judge_whether_insensor(my_d,my_f):               
                pass
            else:                                                                     
                self.delta_p() #delta_poisiton               
                self.drift_v(my_d,my_f) #drift_position                   
                self.drift_s_step(my_d) #drift_next_posiiton
                self.charge_collection(my_f)             
                self.save_inf_track(my_d) 
                self.drift_end_condition()
            self.n_step+=1 
        
        
    def initial_parameter(self):
        """ initial the parameter in the drift """
        self.end_cond=0
        self.d_time=1.0e-9
        self.path_len=0
        self.n_step=0
        self.charge=0

    def judge_whether_insensor(self,my_d,my_f):
        """
        Judge the whether the point(x,y,z) is in the sensor
        and whether the electric field is zero
        """
        # electic field of the position
        self.e_field = my_f.get_e_field(self.d_x,
                                        self.d_y,
                                        self.d_z)
        # modify this part										
        if (self.d_y>=(my_d.l_y-1.0) or self.d_x>=(my_d.l_x-1.0) or self.d_z>=(my_d.l_z-1.0)):
            self.end_cond=3  
        elif (self.d_y<=(1.0) or self.d_x<=(1.0) or self.d_z<=(1.0)):
            self.end_cond=8                    
        elif (self.e_field[0]==0 and self.e_field[1]==0 and self.e_field[1]==0):
            self.end_cond=9
        return self.end_cond

    def delta_p(self):
        """ sstep(1um) split to three position """
        if(self.charg)>0:
            FF=self.list_add(self.e_field,
                             self.cross(self.e_field,self.BB,self.muhh))
        else:
            FF=self.list_sub(self.e_field,
                             self.cross(self.e_field,self.BB,self.muhe))   
        total_ef = self.root_mean_square(FF)
        if(total_ef!=0):
            self.delta_x=-self.sstep*self.charg*FF[0]/total_ef
            self.delta_y=-self.sstep*self.charg*FF[1]/total_ef
            self.delta_z=-self.sstep*self.charg*FF[2]/total_ef
        else:
            self.delta_x=0.0
            self.delta_y=0.0
            self.delta_z=0.0

    def cross(self,p1,p2,scale):
        """ Get vector cross product of p1, p2 """
        o1 = p1[1]*p2[2]-p1[2]*p2[1]
        o2 = p1[2]*p2[0]-p1[0]*p2[2]
        o3 = p1[0]*p2[1]-p1[1]*p2[0]
        return [scale*o1,scale*o2,scale*o3]

    def root_mean_square(self,p1):
        " Return root_mean_square of p1"
        return math.sqrt(p1[0]*p1[0]+p1[1]*p1[1]+p1[2]*p1[2])

    def list_add(self,p1,p2):
        " Return the added two lists. eg:[1,2,3]+[1,2,3] = [2,4,6]"
        return [ a+b for a,b in zip(p1,p2) ]

    def list_sub(self,p1,p2):
        " Return the added two lists. eg:[1,2,3]-[1,2,3] = [0,0,0]"
        return [ a-b for a,b in zip(p1,p2) ]

    def drift_v(self,my_d,my_f):
        """ The drift of e-h pairs at electric field """
        e_delta_f = my_f.get_e_field(self.d_x+self.delta_x,
                                     self.d_y+self.delta_y,
                                     self.d_z+self.delta_z)
        te_delta_f = self.root_mean_square(e_delta_f)
        aver_e = (self.root_mean_square(self.e_field) 
                  + te_delta_f)/2.0*1e4            # V/cm                                                        
        mobility = sic_mobility(self.charg,aver_e,my_d)  # mobility cm2/(V s) v : cm/s
        self.v_drift = mobility*aver_e 
        #drift part
        if(self.v_drift==0):
            self.delta_x=0.0
            self.delta_y=0.0
            self.delta_z=0.0
            self.dif_x=0.0
            self.dif_y=0.0
            self.dif_z=0.0
            self.end_cond=9
        else:
            #off when the field gets large enough
            DiffOffField=100.0  # if the electric field  
                                # > 100V/um, the holes will multiplicat             
            if(te_delta_f < DiffOffField):
                self.s_time = self.sstep*1e-4/self.v_drift
                s_sigma = math.sqrt(2.0*self.kboltz*mobility
                                    *my_d.temperature*self.s_time)
                self.dif_x=random.gauss(0.0,s_sigma)*1e4
                self.dif_y=random.gauss(0.0,s_sigma)*1e4
                self.dif_z=random.gauss(0.0,s_sigma)*1e4                       
            else:
                print("the eletric field is too big, \
                       the multiplication appear. The system shold end. ")
                sys.exit(0)

    def drift_s_step(self,my_d):
        """" Drift distance
        d_x: raw position
        delta_x: eletric field
        dif_x thermal diffusion
        """
        # x axis   
        if((self.d_x+self.delta_x+self.dif_x)>=my_d.l_x): 
            self.d_cx = my_d.l_x
        elif((self.d_x+self.delta_x+self.dif_x)<0):
            self.d_cx = 0
        else:
            self.d_cx = self.d_x+self.delta_x+self.dif_x
        # y axis
        if((self.d_y+self.delta_y+self.dif_y)>=my_d.l_y): 
            self.d_cy = my_d.l_y
        elif((self.d_y+self.delta_y+self.dif_y)<0):
            self.d_cy = 0
        else:
            self.d_cy = self.d_y+self.delta_y+self.dif_y
        # z axis
        if((self.d_z+self.delta_z+self.dif_z)>=my_d.l_z): 
            self.d_cz = my_d.l_z
        elif((self.d_z+self.delta_z+self.dif_z)<0):
            self.d_cz = 0
        else:
            self.d_cz = self.d_z+self.delta_z+self.dif_z

    def charge_collection(self,my_f):
        """ Calculate charge collection """ 
        self.wpot = my_f.get_w_p(self.d_cx,self.d_cy,self.d_cz)
        delta_Uw = (self.wpot 
                    - my_f.get_w_p(self.d_x,self.d_y,self.d_z))
        self.charge=self.charg*delta_Uw
        if(self.v_drift!=0):
            self.d_time=self.d_time+self.sstep*1e-4/self.v_drift
            self.path_len+=self.sstep
        self.d_x=self.d_cx
        self.d_y=self.d_cy
        self.d_z=self.d_cz

    def drift_end_condition(self): 
        """ Judge whether the drift loop should end """
        if(self.wpot>(1-1e-5)):
            self.end_cond=1
        if(self.d_x<=0):
            self.end_cond=2
        if(self.d_y<=0):
            self.end_cond=4
        if(self.d_z<=0):
            self.end_cond=5
        if(self.path_len>self.max_drift_len):
            self.end_cond=6
        if(self.n_step>10000):
            self.end_cond=7

    def save_inf_track(self,my_d):
        """ Save the information in the dictionary """
        if(((self.charge<0 and my_d.v_voltage<0)  
             or (self.charge>0 and my_d.v_voltage>0))): 
            if(self.charg>0):
                self.d_dic_p["tk_"+str(self.n_track)][0].append(self.d_x)
                self.d_dic_p["tk_"+str(self.n_track)][1].append(self.d_y)
                self.d_dic_p["tk_"+str(self.n_track)][2].append(self.d_z)
                self.d_dic_p["tk_"+str(self.n_track)][3].append(self.charge)
                self.d_dic_p["tk_"+str(self.n_track)][4].append(self.d_time)
            else:
                self.d_dic_n["tk_"+str(self.n_track)][0].append(self.d_x)
                self.d_dic_n["tk_"+str(self.n_track)][1].append(self.d_y)
                self.d_dic_n["tk_"+str(self.n_track)][2].append(self.d_z)
                self.d_dic_n["tk_"+str(self.n_track)][3].append(self.charge)
                self.d_dic_n["tk_"+str(self.n_track)][4].append(self.d_time)

    def get_current(self,my_d):
        """ Charge distribution to total current"""
        self.reset_start(my_d)
        test_p = ROOT.TH1F("test+","test+",my_d.n_bin,my_d.t_start,my_d.t_end)
        test_n = ROOT.TH1F("test-","test-",my_d.n_bin,my_d.t_start,my_d.t_end)
        total_pairs=0
        sic_loss_e=self.meter_choose(my_d)
        for j in range(len(self.tracks_p)-2):
            test_p,test_n = self.get_trackspn(my_d, test_p, test_n, j)   
            n_pairs=self.tracks_step_edep[j]*1e6/sic_loss_e
            total_pairs+=n_pairs
            test_p.Scale(n_pairs)
            test_n.Scale(n_pairs)            
            my_d.positive_cu.Add(test_p)
            my_d.negative_cu.Add(test_n)
            test_p.Reset()
            test_n.Reset()

        self.landau_t_pairs = self.tracks_t_edep*1e6/sic_loss_e
        print("landau_t_pairs=%s"%self.landau_t_pairs)

        if total_pairs != 0:
            n_scale = self.landau_t_pairs/total_pairs
        else:
            n_scale=0
        my_d.sum_cu.Add(my_d.positive_cu)
        my_d.sum_cu.Add(my_d.negative_cu)
        my_d.sum_cu.Scale(n_scale)

    def reset_start(self,my_d):
        """ Reset th1f """
        my_d.positive_cu.Reset()
        my_d.negative_cu.Reset()
        my_d.sum_cu.Reset()
        self.sum_p_current = []

    def meter_choose(self,my_d):
        """ Judge the material of sensor """
        if (my_d.mater == 1): # silicon carbide
            sic_loss_e = 8.4 #ev
        elif (my_d.mater == 0):   # silicon
            sic_loss_e = 3.6 #ev
        return sic_loss_e

    def get_trackspn(self, my_d, test_p, test_n, j):
        """ Total current of each e-h pair"""
        e0 = 1.60217733e-19
        for i in range(len(self.d_dic_p["tk_"+str(j+1)][2])):
            test_p.Fill(self.d_dic_p["tk_"+str(j+1)][4][i],
                        self.d_dic_p["tk_"+str(j+1)][3][i]/my_d.t_bin*e0)   
        for i in range(len(self.d_dic_n["tk_"+str(j+1)][2])):
            test_n.Fill(self.d_dic_n["tk_"+str(j+1)][4][i],
                        self.d_dic_n["tk_"+str(j+1)][3][i]/my_d.t_bin*e0)
        return test_p, test_n
        
# # # mobility model
def sic_mobility(charge,aver_e,my_d):
    """ SiC mobility model 
    SiC reference: TCAD SIMULATION FOR ALPHA-PARTICLE SPECTROSCOPY USING SIC SCHOTTKY DIODE
    Si  reference: Signal development in irradiated silicon detectors
    """
    T=my_d.temperature
    E=aver_e
    Neff=abs(my_d.d_neff)
    #silicon
    if my_d.mater == 0:
        alpha = 0.72*math.pow(T/300.0,0.065)
        if(charge>0):
            ulp = 460.0 * math.pow(T / 300.0, -2.18)
            uminp = 45.0*math.pow(T / 300.0, -0.45)
            Crefp = 2.23e17*math.pow(T / 300.0, 3.2)
            betap = 1.0
            vsatp = 9.05e6 * math.sqrt(math.tanh(312.0/T))
            lfm = uminp + (ulp-uminp)/(1.0 + math.pow(Neff*1e12 / Crefp, alpha))
            hfm = 2*lfm / (1.0+math.pow(1.0 + math.pow(2*lfm * E / vsatp, betap), 1.0 / betap))                        
        else:
            uln = 1430.0 * math.pow(T / 300.0, -2.0)
            uminn = 80.0*math.pow(T / 300.0, -0.45)
            Crefn = 1.12e17*math.pow(T/300.0,3.2)
            betan = 2
            vsatn = 1.45e7 * math.sqrt(math.tanh(155.0/T))
            lfm = uminn + (uln-uminn)/ (1.0 + math.pow(Neff*1e12 / Crefn, alpha))
            hfm = 2*lfm / (1.0+math.pow(1.0 + math.pow(2*lfm * E / vsatn, betan), 1.0/betan))
    #silicon carbide
    elif my_d.mater == 1:
        if(charge>0):
            alpha = 0.34
            ulp = 124.0 * math.pow(T / 300.0, -2.0)
            uminp = 15.9
            Crefp = 1.76e19
            betap = 1.213 * math.pow(T / 300.0, 0.17)
            vsatp = 2e7 * math.pow(T / 300.0, 0.52)
            lfm = uminp + ulp/(1.0 + math.pow(Neff*1e12 / Crefp, alpha))
            hfm = lfm / (math.pow(1.0 + math.pow(lfm * E / vsatp, betap), 1.0 / betap))                        
        else:
            alpha = 0.61
            ulp = 947.0 * math.pow(T / 300.0, -2)
            Crefp = 1.94e19
            betap = 1.0 * math.pow(T / 300.0, 0.66)
            vsatp = 2e7 * math.pow(T / 300.0, 0.87)
            lfm = ulp/ (1.0 + math.pow(Neff*1e12 / Crefp, alpha))
            hfm = lfm / (math.pow(1.0 + math.pow(lfm * E / vsatp, betap), 1.0/betap))                      
    return hfm


class CalCurrent2D:

    def __init__(self,track,fen,det,model="lgad2D"):

        self.track = track
        self.fen = fen
        self.det = det

        # initial tracks
        self.delta_track_info_dic_n = {}
        self.delta_track_info_dic_p = {}

        # gain tracks
        # self.delta_gain_track_info_dic_n_n = {}
        # self.delta_gain_track_info_dic_n_p = {}
        # self.delta_gain_track_info_dic_p_n = {}
        # self.delta_gain_track_info_dic_p_p = {}
        self.delta_gain_track_info_dic = {}

        self.track_time = 0.
        self.track_x = 0.
        self.track_y = 0.
        self.track_charges = 0.
        self.track_current = 0.

        self.end_condition = 0

        self.delta_x=0.
        self.delta_y=0.
        self.dif_x=0.
        self.dif_y=0.

        self.s_time = 0.

        self.s_gain = 1.

        self.path_len=0
        self.n_step=0

        for n in range(len(track.track_position)):

            # initial tracks
            self.delta_track_info_dic_n["tk_"+str(n+1)] = [ [] for n in range(6) ]   # track_time, track_x, track_y, track_charge, track_current, track_gain
            self.delta_track_info_dic_p["tk_"+str(n+1)] = [ [] for n in range(6) ] 

        self.cal_current(model)
        self.draw_drift_path(det)
        if isinstance(track,MIPs):
            self.landau_scale()
            self.draw_drift_path(det)

    
    def drift_diffusion(self,det,fen):

        my_mobility = Mobility(det.mat_name)

        self.muhh=1650   # mobility related with the magnetic field (now silicon useless)
        self.muhe=310
        self.BB=np.array([0,0])
        if det.det_width > 100:
            self.sstep = 1 #drift step
        else:
            self.sstep=0.1 # drift step
        self.kboltz=8.617385e-5 # eV/K
        self.max_drift_len=1e9 # maximum driftlength [um]

        self.delta_x=0.
        self.delta_y=0. 
        

        #
        # drift
        #

        # magnetic field effect

        # for hole
        if(self.charge)>0:
            FF=self.e_field+self.muhh*np.cross(self.e_field,self.BB)
        # for electron    
        else:
            FF=self.e_field-self.muhe*np.cross(self.e_field,self.BB)

        # the delta x with electric field and magnetic field for unit e or h

        if(np.linalg.norm(FF)!=0):
            self.delta_x = self.sstep*self.charge*FF[0]/np.linalg.norm(FF)
            self.delta_y = self.sstep*self.charge*FF[1]/np.linalg.norm(FF)
        else:
            self.delta_x=0
            self.delta_y=0 

        # cal current step ef & wef
        efx = fen.cal_point_field(self.track_x+self.delta_x, self.track_y+self.delta_y,fen.electric_field_x_value)
        efy = fen.cal_point_field(self.track_x+self.delta_x, self.track_y+self.delta_y,fen.electric_field_y_value)
        ef = np.array([efx,efy])

        ef_value = np.linalg.norm(ef)*1e4 #V/cm

        wefx = fen.cal_point_field(self.track_x+self.delta_x, self.track_y+self.delta_y,fen.weighting_electric_field_x_value)
        wefy = fen.cal_point_field(self.track_x+self.delta_x, self.track_y+self.delta_y,fen.weighting_electric_field_y_value)
        wef = np.array([wefx,wefy])

        self.wef_value = np.linalg.norm(wef)


        pos = [self.track_x+self.delta_x, self.track_y+self.delta_y]

        self.drift_velocity = my_mobility.cal_mobility(det, pos, self.charge, ef_value)*ef_value
        # print(self.drift_velocity)

        self.e_field = ef
        self.we_field = wef

        #
        # diffusion
        #
        if(self.drift_velocity == 0):
            self.delta_x=0
            self.delta_y=0
            self.dif_x=0
            self.dif_y=0
            self.end_condition = 9
        else:
            if(det.mat_name=='Si'):
                DiffOffField=8*1e4 #V/cm
            if(det.mat_name=='SiC'):
                DiffOffField=100*1e4 #V/cm
                
            if(ef_value<DiffOffField):
                self.s_time=self.sstep*1e-4/self.drift_velocity
                s_sigma= math.sqrt(2*self.kboltz*my_mobility.cal_mobility(det, pos, self.charge, ef_value)*det.temperature*self.s_time)
                self.dif_x=random.gauss(0,s_sigma)*1e4
                self.dif_y=random.gauss(0,s_sigma)*1e4


            else:
                self.dif_x=0.0
                self.dif_y=0.0 

        #
        # multiplication
        #
        my_avalanche = Avalanche(det.par_dict['Avalanche'])
        tmp_coefficient = my_avalanche.cal_coefficient(ef_value,self.charge,det.temperature)

        self.s_gain = math.exp(self.sstep*1e-4*tmp_coefficient)


    def update_track_info(self):

        if(self.charge>0):
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][0].append(self.track_time)
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][1].append(self.track_x)
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][2].append(self.track_y)
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][3].append(self.track_charges)
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][4].append(self.track_current)
            self.delta_track_info_dic_p["tk_"+str(self.track_number)][5].append(self.track_gain)
 
        if(self.charge<0):
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][0].append(self.track_time)
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][1].append(self.track_x)
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][2].append(self.track_y)
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][3].append(self.track_charges)
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][4].append(self.track_current)
            self.delta_track_info_dic_n["tk_"+str(self.track_number)][5].append(self.track_gain)

    def update_gain_track_info(self):
        self.delta_gain_track_info_dic[self.track_name][0].append(self.track_time)
        self.delta_gain_track_info_dic[self.track_name][1].append(self.track_x)
        self.delta_gain_track_info_dic[self.track_name][2].append(self.track_y)
        self.delta_gain_track_info_dic[self.track_name][3].append(self.track_charges)
        self.delta_gain_track_info_dic[self.track_name][4].append(self.track_current)

    def update_step(self,det):
        self.track_time = self.track_time + self.s_time
        self.track_x = self.track_x + self.delta_x + self.dif_x
        self.track_y = self.track_y + self.delta_y + self.dif_y
        self.track_charges = self.charge
        self.track_gain *= self.s_gain

        if(self.track_x>=det.det_width):
            self.track_x = det.det_width
        
        if(self.track_x<0):
            self.track_x = 0

        if(self.track_y>=det.det_thin):
            self.track_y = det.det_thin
        
        if(self.track_y<0):
            self.track_y = 0
        
    def update_end_condition(self):

        if(self.we_field.all()>(1-1e-5)):
            self.end_condition=1
        if(self.track_x<=0):
            self.end_condition=2
        if(self.track_y<=0):
            self.end_condition=3
        else: 
            self.end_condition=0
            
        if(self.path_len>self.max_drift_len):
            self.end_condition=5
        if(self.n_step>10000):
            self.end_condition=7
        

    def cal_current(self,model):

        e0 = 1.60217733e-19
        track = self.track
        fen = self.fen
        det = self.det
        #
        # initial carrier track
        #
        
        for i in range(len(track.track_position)):

            if track.ionized_pairs[i]<=0:
                continue

            self.track_number = i+1

            for j in range(2):

                if(j==0):
                    self.charge = 1
                    #self.charges=1*track.ionized_pairs[i] # hole

                if(j==1):
                    self.charge = -1
                    #self.charges=-1*track.ionized_pairs[i] # electron
                
                self.track_time = 0.
                self.track_x = track.track_position[i][0]
                self.track_y = track.track_position[i][1]
                self.track_charges = 0.
                self.track_current = 0.
                self.track_gain = 1.

                self.end_condition = 0
                while(self.end_condition==0):

                    if(self.track_y>=(det.det_thin-1) or self.track_x>=(det.det_width-1)):
                         
                        self.end_condition=4

                    else:
                        efx = fen.cal_point_field(self.track_x, self.track_y,fen.electric_field_x_value)
                        efy = fen.cal_point_field(self.track_x, self.track_y,fen.electric_field_y_value)

                        ef = np.array([efx,efy])
                        ef_value = np.linalg.norm(ef)*1e4

                        self.e_field = np.array([efx,efy])

                        wefx = fen.cal_point_field(self.track_x, self.track_y,fen.weighting_electric_field_x_value)
                        wefy = fen.cal_point_field(self.track_x, self.track_y,fen.weighting_electric_field_y_value)

                        wef = np.array([wefx,wefy])
                        wef_value = np.linalg.norm(wef)*1e4

                        self.we_field = np.array([wefx,wefy])                                

                        self.drift_diffusion(det,fen)

                        if (self.drift_velocity!=0):
                            self.path_len+=self.sstep
                        # SR current

                        self.track_current = abs(self.charge*e0*self.drift_velocity*wef_value)

                        self.update_track_info()
                        self.update_step(det)
                        self.update_end_condition()
                    self.n_step +=1

        det.positive_cu.Reset()
        det.negative_cu.Reset()
        det.sum_cu.Reset()

        temp_positive_cu = ROOT.TH1F("temp+","temp+",det.n_bin,0,det.t_end)
        temp_negative_cu = ROOT.TH1F("temp-","temp-",det.n_bin,0,det.t_end)
        temp_sum_cu = ROOT.TH1F("temp_sum","temp_sum",det.n_bin,0,det.t_end)
        
        #
        # initial current
        #
        for i in range(len(track.track_position)):

            for j in range(len(self.delta_track_info_dic_p["tk_"+str(i+1)][0])):
                temp_positive_cu.Fill(self.delta_track_info_dic_p["tk_"+str(i+1)][0][j], self.delta_track_info_dic_p["tk_"+str(i+1)][4][j])
            temp_positive_cu.Scale(track.ionized_pairs[i])

            for k in range(len(self.delta_track_info_dic_n["tk_"+str(i+1)][0])):
                temp_negative_cu.Fill(self.delta_track_info_dic_n["tk_"+str(i+1)][0][k], self.delta_track_info_dic_n["tk_"+str(i+1)][4][k])
            temp_negative_cu.Scale(track.ionized_pairs[i])

            det.positive_cu.Add(temp_positive_cu)
            det.negative_cu.Add(temp_negative_cu)

            temp_positive_cu.Reset()
            temp_negative_cu.Reset()

        det.sum_cu.Add(det.positive_cu)
        det.sum_cu.Add(det.negative_cu)

        if model == "lgad2D":
            #
            # gain carrier track
            #

            # get gain tracks start info
            self.gain_track_info_list = [] #[[name,time,x,y,charge,pairs]]

            for i in range(len(track.track_position)):

                for j in range(len(self.delta_track_info_dic_p["tk_"+str(i+1)][0])):
                    
                    if(self.delta_track_info_dic_p["tk_"+str(i+1)][5][j]>1.0):

                        tmp_gain_time = self.delta_track_info_dic_p["tk_"+str(i+1)][0][j]
                        tmp_gain_x = self.delta_track_info_dic_p["tk_"+str(i+1)][1][j]
                        tmp_gain_y = self.delta_track_info_dic_p["tk_"+str(i+1)][2][j]
                        tmp_gain_pairs = abs(self.delta_track_info_dic_p["tk_"+str(i+1)][3][j])
                        tmp_pairs = track.ionized_pairs[i]*(np.max(self.delta_track_info_dic_n["tk_"+str(i+1)][5])-1)
                        tmp_gain_current = 0.

                        self.gain_track_info_list.append(["tk_"+str(i+1)+"_p_n",tmp_gain_time,tmp_gain_x,tmp_gain_y,-tmp_gain_pairs,tmp_pairs])

                        self.gain_track_info_list.append(["tk_"+str(i+1)+"_p_p",tmp_gain_time,tmp_gain_x,tmp_gain_y,tmp_gain_pairs,tmp_pairs])

                        self.delta_gain_track_info_dic["tk_"+str(i+1)+"_p_n"] = [ [] for n in range(5) ]
                        self.delta_gain_track_info_dic["tk_"+str(i+1)+"_p_p"] = [ [] for n in range(5) ]

                        break

                for k in range(len(self.delta_track_info_dic_n["tk_"+str(i+1)][0])):

                    if(self.delta_track_info_dic_n["tk_"+str(i+1)][5][k]>1.0):

                        tmp_gain_time = self.delta_track_info_dic_n["tk_"+str(i+1)][0][k]
                        tmp_gain_x = self.delta_track_info_dic_n["tk_"+str(i+1)][1][k]
                        tmp_gain_y = self.delta_track_info_dic_n["tk_"+str(i+1)][2][k]
                        tmp_gain_pairs = abs(self.delta_track_info_dic_n["tk_"+str(i+1)][3][k])
                        tmp_pairs = track.ionized_pairs[i]*(np.max(self.delta_track_info_dic_n["tk_"+str(i+1)][5])-1)
                        tmp_gain_current = 0.

                        self.gain_track_info_list.append(["tk_"+str(i+1)+"_n_n",tmp_gain_time,tmp_gain_x,tmp_gain_y,-tmp_gain_pairs,tmp_pairs])

                        self.gain_track_info_list.append(["tk_"+str(i+1)+"_n_p",tmp_gain_time,tmp_gain_x,tmp_gain_y,tmp_gain_pairs,tmp_pairs])

                        self.delta_gain_track_info_dic["tk_"+str(i+1)+"_n_n"] = [ [] for n in range(5) ]
                        self.delta_gain_track_info_dic["tk_"+str(i+1)+"_n_p"] = [ [] for n in range(5) ]

                        break

            #
            # cal gain current
            #

            for i in range(len(self.gain_track_info_list)):

                self.charge = self.gain_track_info_list[i][4]

                self.track_name = self.gain_track_info_list[i][0]
                self.track_time = self.gain_track_info_list[i][1]
                self.track_x = self.gain_track_info_list[i][2]
                self.track_y = self.gain_track_info_list[i][3]
                self.track_charges = self.gain_track_info_list[i][4]
                self.track_current = 0.

                self.n_step = 0
                self.path_len = 0
                self.end_condition = 0
                while(self.end_condition==0):

                    if(self.track_y>=(det.det_thin-1) or self.track_x>=(det.det_width-1)):
                        
                        self.end_condition=4

                    else:
                        efx = fen.cal_point_field(self.track_x, self.track_y,fen.electric_field_x_value)
                        efy = fen.cal_point_field(self.track_x, self.track_y,fen.electric_field_y_value)
                        ef = np.array([efx,efy])
                        ef_value = np.linalg.norm(ef)*1e4
                        self.e_field = np.array([efx,efy])
                        wefx = fen.cal_point_field(self.track_x, self.track_y,fen.weighting_electric_field_x_value)
                        wefy = fen.cal_point_field(self.track_x, self.track_y,fen.weighting_electric_field_y_value)
                        wef = np.array([wefx,wefy])
                        wef_value = np.linalg.norm(wef)*1e4
                        self.we_field = np.array([wefx,wefy])
                                                    
                        self.drift_diffusion(det,fen)

                        if (self.drift_velocity!=0):
                            self.path_len+=self.sstep
                        # SR current
                        self.track_current = abs(self.charge*e0*self.drift_velocity*wef_value)

                        self.update_gain_track_info()
                        self.update_step(det)
                        self.update_end_condition()
                    self.n_step+=1

            det.gain_positive_cu.Reset()
            det.gain_negative_cu.Reset()

            det.gain_n_n_cu.Reset()
            det.gain_n_p_cu.Reset()
            det.gain_p_n_cu.Reset()
            det.gain_p_p_cu.Reset()

            #
            # gain current
            #

            temp_gain_cu = ROOT.TH1F("temp_gain","temp_gain",det.n_bin,0,det.t_end)

            for i in range(len(self.gain_track_info_list)):

                tmp_track_name = self.gain_track_info_list[i][0]

                for j in range(len(self.delta_gain_track_info_dic[tmp_track_name][0])):
                    
                    temp_gain_cu.Fill(self.delta_gain_track_info_dic[tmp_track_name][0][j],self.delta_gain_track_info_dic[tmp_track_name][4][j])
                temp_gain_cu.Scale(self.gain_track_info_list[i][5])

                if(tmp_track_name[-1]=='p'):
                    det.gain_positive_cu.Add(temp_gain_cu)

                if(tmp_track_name[-1]=='n'):
                    det.gain_negative_cu.Add(temp_gain_cu)


                if(tmp_track_name[-3:]=='n_n'):
                    det.gain_n_n_cu.Add(temp_gain_cu)

                if(tmp_track_name[-3:]=='n_p'):
                    det.gain_n_p_cu.Add(temp_gain_cu)

                if(tmp_track_name[-3:]=='p_n'):
                    det.gain_p_n_cu.Add(temp_gain_cu)
                    
                if(tmp_track_name[-3:]=='p_p'):
                    det.gain_p_p_cu.Add(temp_gain_cu)


                temp_gain_cu.Reset()

            det.sum_cu.Add(det.gain_positive_cu)
            det.sum_cu.Add(det.gain_negative_cu)

    def landau_scale(self):
        #landau_pairs = track.mips_landau()
        landau_pairs = 4000

        current_scale = landau_pairs/self.track.ionized_total_pairs

        self.det.positive_cu.Scale(current_scale)
        self.det.negative_cu.Scale(current_scale)
        self.det.gain_positive_cu.Scale(current_scale)
        self.det.gain_negative_cu.Scale(current_scale)

        self.det.gain_n_n_cu.Scale(current_scale)
        self.det.gain_n_p_cu.Scale(current_scale)
        self.det.gain_p_n_cu.Scale(current_scale)
        self.det.gain_p_p_cu.Scale(current_scale)
        self.det.sum_cu.Scale(current_scale)
        
    def draw_drift_path(self,det):
        # ROOT.gStyle.SetOptStat(0)

        c1 = ROOT.TCanvas("c1", "canvas1", 200,10,1000, 500)
        c1.Divide(2,1)

        x_array=array('f')
        y_array=array('f')

        mg1 = ROOT.TMultiGraph("mg1","initial current path")
        for i in range(len(self.delta_track_info_dic_p)):
            n=len(self.delta_track_info_dic_p["tk_"+str(i+1)][0])
            if(n>0):
                x_array.extend(self.delta_track_info_dic_p["tk_"+str(i+1)][1])
                y_array.extend(self.delta_track_info_dic_p["tk_"+str(i+1)][2])             
                gr_p = ROOT.TGraph(n,x_array,y_array)
                gr_p.SetMarkerColor(4)
                gr_p.SetLineColor(4)
                gr_p.SetLineStyle(1)
                mg1.Add(gr_p)
                del x_array[:]
                del y_array[:]
        for j in range(len(self.delta_track_info_dic_n)):
            m=len(self.delta_track_info_dic_n["tk_"+str(j+1)][0])
            if(m>0):
                x_array.extend(self.delta_track_info_dic_n["tk_"+str(j+1)][1])
                y_array.extend(self.delta_track_info_dic_n["tk_"+str(j+1)][2])             
                gr_n = ROOT.TGraph(m,x_array,y_array)
                gr_n.SetMarkerColor(2)
                gr_n.SetLineColor(2)
                gr_n.SetLineStyle(1)
                mg1.Add(gr_n)
                del x_array[:]
                del y_array[:]
        mg1.GetXaxis().SetLimits(0,det.det_width)
        mg1.GetYaxis().SetLimits(0,det.det_thin)
        c1.cd(1)
        mg1.Draw("APL")

        mg2 = ROOT.TMultiGraph("mg2","gain current path")

        for j in range(len(self.delta_gain_track_info_dic)):

            track_name = self.gain_track_info_list[j][0]
            n = len(self.delta_gain_track_info_dic[track_name][0])
            
            if(n>0):
                x_array.extend(self.delta_gain_track_info_dic[track_name][1])
                y_array.extend(self.delta_gain_track_info_dic[track_name][2])

                if(track_name[-1]=='p'):
                    gr_gain_p = ROOT.TGraph(n,x_array,y_array)
                    gr_gain_p.SetMarkerColor(4)
                    gr_gain_p.SetLineColor(4)
                    gr_gain_p.SetLineStyle(1)
                    mg2.Add(gr_gain_p)
                    del x_array[:]
                    del y_array[:]
                
                if(track_name[-1]=='n'):
                    gr_gain_n = ROOT.TGraph(n,x_array,y_array)
                    gr_gain_n.SetMarkerColor(2)
                    gr_gain_n.SetLineColor(2)
                    gr_gain_n.SetLineStyle(1)
                    mg2.Add(gr_gain_n)
                    del x_array[:]
                    del y_array[:]
        mg2.GetXaxis().SetLimits(0,det.det_width)
        mg2.GetYaxis().SetLimits(0,det.det_thin)
        c1.cd(2)
        mg2.Draw("APL")

        c1.SaveAs("./fig/%s_%s_drift_path_%dV.pdf"%(det.mat_name,det.det_model,det.bias_voltage))



# -*- encoding: utf-8 -*-
'''
Description: 
    geat4_pybind simulate the energy depostion of 
    beta-source time resoltion experiment     
@Date       : 2021/09/02 12:46:27
@Author     : tanyuhang
@version    : 1.0
'''

import geant4_pybind as g4b
import sys
import numpy as np

# Geant4 main process
class Particles:
    def __init__(self, my_d, my_f, dset):
        """
        Description:
            Geant4 main process
            Simulate s_num electrons through device
            Record the energy depositon in the device
        Parameters:
        ---------
        energy_steps : list
            Energy deposition of each step in simulation
        edep_devices : list
            Total energy deposition in device          
        @Modify:
        ---------
            2021/09/02
        """	
        g4_dic = dset.pygeant4
        my_g4d = MyDetectorConstruction(my_d,my_f,g4_dic['name'],g4_dic['maxstep'])		
        if g4_dic['g4_vis']: 
            ui = None
            ui = g4b.G4UIExecutive(len(sys.argv), sys.argv)
        gRunManager = g4b.G4RunManagerFactory.CreateRunManager(g4b.G4RunManagerType.Default)
        rand_engine= g4b.RanecuEngine()
        g4b.HepRandom.setTheEngine(rand_engine)
        g4b.HepRandom.setTheSeed(dset.g4seed)
        gRunManager.SetUserInitialization(my_g4d)	
        # set physics list
        physics_list =  g4b.FTFP_BERT()
        physics_list.SetVerboseLevel(1)
        physics_list.RegisterPhysics(g4b.G4StepLimiterPhysics())
        gRunManager.SetUserInitialization(physics_list)
        # define golbal parameter
        global s_eventIDs,s_edep_devices,s_p_steps,s_energy_steps,s_events_angle
        s_eventIDs,s_edep_devices,s_p_steps,s_energy_steps,s_events_angle=[],[],[],[],[]

        #define action
        gRunManager.SetUserInitialization(MyActionInitialization(
                                          g4_dic['par_in'],
                                          g4_dic['par_out']))
        if g4_dic['g4_vis']:    
            visManager = g4b.G4VisExecutive()
            visManager.Initialize()
            UImanager = g4b.G4UImanager.GetUIpointer()
            UImanager.ApplyCommand('/control/execute init_vis.mac')
        else:
            UImanager = g4b.G4UImanager.GetUIpointer()
            UImanager.ApplyCommand('/run/initialize')
            
        gRunManager.BeamOn(int(dset.total_events))
        if g4_dic['g4_vis']:  
            ui.SessionStart()
        self.p_steps=s_p_steps
        self.energy_steps=s_energy_steps
        self.edep_devices=s_edep_devices
        self.events_angle=s_events_angle
        self.init_tz_device = my_g4d.init_tz_device
        del s_eventIDs,s_edep_devices,s_p_steps,s_energy_steps,s_events_angle
        
    def __del__(self):
        pass

#Geant4 for particle drift path
class MyDetectorConstruction(g4b.G4VUserDetectorConstruction):                
    "My Detector Construction"
    def __init__(self,my_d,my_f,sensor_model,maxStep=0.5):
        g4b.G4VUserDetectorConstruction.__init__(self)
        self.solid = {}
        self.logical = {}
        self.physical = {}
        self.checkOverlaps = True
        self.create_world(my_d)
        #3D source order: beta->sic->si
        #2D source order: beta->Si->SiC
        tx_all = my_d.l_x/2.0*g4b.um
        ty_all = my_d.l_y/2.0*g4b.um
        if "plugin3D" in sensor_model:
            tz_Si = 0*g4b.um
            tz_device = 10000*g4b.um+my_d.l_z/2.0*g4b.um
            self.init_tz_device = 10000
            tz_pcb2 = 10000*g4b.um-750*g4b.um
            device_x = (my_f.sx_r-my_f.sx_l)*g4b.um 
            device_y = (my_f.sy_r-my_f.sy_l)*g4b.um
            device_z = my_d.l_z*g4b.um
        elif "planar3D" in sensor_model:
            tz_Si = 10000*g4b.um
            tz_device = my_d.l_z/2.0*g4b.um
            self.init_tz_device = 0
            tz_pcb2 = -1100*g4b.um
            device_x = my_d.l_x*g4b.um 
            device_y = my_d.l_y*g4b.um
            device_z = my_d.l_z*g4b.um
        self.create_AlorSi_box(
                                name = "Al",
                                sidex = my_d.l_x*g4b.um,
                                sidey = my_d.l_y*g4b.um,
                                sidez = 10*g4b.um,
                                translation = [tx_all,ty_all,15000*g4b.um],
                                material_type = "G4_Al",
                                colour = [1,0.1,0.8],
                                mother = 'world')
        self.create_AlorSi_box(
                                name = "Si_main",
                                sidex = 1300*g4b.um,
                                sidey = 1300*g4b.um,
                                sidez = 33*g4b.um,
                                translation = [tx_all,ty_all,tz_Si],
                                material_type = "G4_Si",
                                colour = [1,1,1],
                                mother = 'world')
        self.create_AlorSi_box(
                                name = "Si_sub",
                                sidex = 1300*g4b.um,
                                sidey = 1300*g4b.um,
                                sidez = 300*g4b.um,
                                translation = [tx_all,ty_all,
                                               tz_Si-166.5*g4b.um],
                                material_type = "G4_Si",
                                colour = [0,0,1],
                                mother = 'world')
        self.create_pcb_board(
                                name = "pcb1",
                                sidex = 20000*g4b.um,
                                sidey = 20000*g4b.um,
                                sidez = 1500*g4b.um,
                                translation = [tx_all,ty_all,
                                               tz_Si-1066.5*g4b.um],
                                tub_radius = 500*g4b.um,
                                tub_depth = 1500*g4b.um,
                                material_Si = "Si",
                                material_O = "O",
                                colour = [0,0.5,0.8],   
                                mother = 'world')

        self.create_sic_box(
                            name = "Device",
                            sidex = device_x,
                            sidey = device_y,
                            sidez = device_z,
                            translation = [tx_all,ty_all,tz_device],
                            material_Si = "Si",
                            material_c = "C",
                            colour = [1,0,0],
                            mother = 'world')

        self.create_pcb_board(
                                name = "pcb2",
                                sidex = 20000*g4b.um,
                                sidey = 20000*g4b.um,
                                sidez = 1500*g4b.um,
                                translation = [tx_all,ty_all,tz_pcb2],
                                tub_radius = 500*g4b.um,
                                tub_depth = 1500*g4b.um,
                                material_Si = "Si",
                                material_O = "O",
                                colour = [0,0.5,0.8],   
                                mother = 'world')

        if "plugin3D" in sensor_model:
            self.create_sic_box(
                                name = "SiC_sub",
                                sidex = my_d.l_x*g4b.um,
                                sidey = my_d.l_y*g4b.um,
                                sidez = 350.0*g4b.um,
                                translation = [tx_all,ty_all,-175.0*g4b.um],
                                material_Si = "Si",
                                material_c = "C",
                                colour = [0,1,1],
                                mother = 'world')
        self.maxStep = maxStep*g4b.um
        self.fStepLimit = g4b.G4UserLimits(self.maxStep)
        self.logical["Device"].SetUserLimits(self.fStepLimit)

    def create_world(self,my_d):

        self.nist = g4b.G4NistManager.Instance()
        material = self.nist.FindOrBuildMaterial("G4_AIR")  
        self.solid['world'] = g4b.G4Box("world",
                                        25000*g4b.um,
                                        25000*g4b.um,
                                        25000*g4b.um)
        self.logical['world'] = g4b.G4LogicalVolume(self.solid['world'], 
                                                    material, 
                                                    "world")
        self.physical['world'] = g4b.G4PVPlacement(None, 
                                                   g4b.G4ThreeVector(0,0,0), 
                                                   self.logical['world'], 
                                                   "world", None, False, 
                                                   0,self.checkOverlaps)
        visual = g4b.G4VisAttributes()
        visual.SetVisibility(False)
        self.logical['world'].SetVisAttributes(visual)

    def create_sic_box(self, **kwargs):
        name = kwargs['name']
        material_si = self.nist.FindOrBuildElement(kwargs['material_Si'],False)
        material_c = self.nist.FindOrBuildElement(kwargs['material_c'],False)
        sic_density = 3.2*g4b.g/g4b.cm3
        SiC = g4b.G4Material("SiC",sic_density,2) 
        SiC.AddElement(material_si,50*g4b.perCent)
        SiC.AddElement(material_c,50*g4b.perCent)
        translation = g4b.G4ThreeVector(*kwargs['translation'])
        visual = g4b.G4VisAttributes(g4b.G4Color(*kwargs['colour']))
        mother = self.physical[kwargs['mother']]
        sidex = kwargs['sidex']
        sidey = kwargs['sidey']
        sidez = kwargs['sidez']

        self.solid[name] = g4b.G4Box(name, sidex/2., sidey/2., sidez/2.)
        
        self.logical[name] = g4b.G4LogicalVolume(self.solid[name], 
                                                 SiC, 
                                                 name)
        self.physical[name] = g4b.G4PVPlacement(None,translation,                                                
                                                name,self.logical[name],
                                                mother, False, 
                                                0,self.checkOverlaps)
        self.logical[name].SetVisAttributes(visual)

    def create_AlorSi_box(self, **kwargs):
        name = kwargs['name']
        material_type = self.nist.FindOrBuildMaterial(kwargs['material_type'],
                                                      False)

        translation = g4b.G4ThreeVector(*kwargs['translation'])
        visual = g4b.G4VisAttributes(g4b.G4Color(*kwargs['colour']))
        mother = self.physical[kwargs['mother']]
        sidex = kwargs['sidex']
        sidey = kwargs['sidey']
        sidez = kwargs['sidez']

        self.solid[name] = g4b.G4Box(name, sidex/2., sidey/2., sidez/2.)
        
        self.logical[name] = g4b.G4LogicalVolume(self.solid[name], 
                                                 material_type, 
                                                 name)
        self.physical[name] = g4b.G4PVPlacement(None,translation,                                                
                                                name,self.logical[name],
                                                mother, False, 
                                                0,self.checkOverlaps)
        self.logical[name].SetVisAttributes(visual)     

    def create_pcb_board(self, **kwargs):
        name = kwargs['name']
        material_si = self.nist.FindOrBuildElement(kwargs['material_Si'],False)
        material_O = self.nist.FindOrBuildElement(kwargs['material_O'],False)
        sic_density = 2.2*g4b.g/g4b.cm3
        SiO2 = g4b.G4Material("SiO2",sic_density,2) 
        SiO2.AddElement(material_si,1)
        SiO2.AddElement(material_O,2)
        translation = g4b.G4ThreeVector(*kwargs['translation'])
        visual = g4b.G4VisAttributes(g4b.G4Color(*kwargs['colour']))
        mother = self.physical[kwargs['mother']]
        sidex = kwargs['sidex']
        sidey = kwargs['sidey']
        sidez = kwargs['sidez']
        tub_radius = kwargs['tub_radius']
        tub_depth = kwargs['tub_depth']

        self.solid[name+"box"] = g4b.G4Box(name+"box", 
                                           sidex/2., sidey/2., sidez/2.)
        self.solid[name+"tub"] = g4b.G4Tubs(name+"tub", 0,tub_radius,
                                            tub_depth, 0,360*g4b.deg)
        self.solid[name] = g4b.G4SubtractionSolid(name,
                                                  self.solid[name+"box"],
                                                  self.solid[name+"tub"])
        self.logical[name] = g4b.G4LogicalVolume(self.solid[name], 
                                                 SiO2, 
                                                 name)
        self.physical[name] = g4b.G4PVPlacement(None,translation,                                                
                                                name,self.logical[name],
                                                mother, False, 
                                                0,self.checkOverlaps)
        self.logical[name].SetVisAttributes(visual)

    def Construct(self): # return the world volume
        self.fStepLimit.SetMaxAllowedStep(self.maxStep)
        return self.physical['world']

    def __del__(self):
        print("use __del__ to delete the MyDetectorConstruction class ")


class MyPrimaryGeneratorAction(g4b.G4VUserPrimaryGeneratorAction):
    "My Primary Generator Action"
    def __init__(self,par_in,par_out):
        g4b.G4VUserPrimaryGeneratorAction.__init__(self)
        par_direction = [ par_out[i] - par_in[i] for i in range(3) ]  
        particle_table = g4b.G4ParticleTable.GetParticleTable()
        electron = particle_table.FindParticle("e-") # define the beta electron
        beam = g4b.G4ParticleGun(1)
        beam.SetParticleEnergy(2.28*g4b.MeV)
        # beam.SetParticleEnergy(0.546*g4b.MeV)
        beam.SetParticleMomentumDirection(g4b.G4ThreeVector(par_direction[0],
                                                            par_direction[1],
                                                            par_direction[2]))
        beam.SetParticleDefinition(electron)
        beam.SetParticlePosition(g4b.G4ThreeVector(par_in[0]*g4b.um,
                                                   par_in[1]*g4b.um,
                                                   par_in[2]*g4b.um))  

        beam2 = g4b.G4ParticleGun(1)
        beam2.SetParticleEnergy(0.546*g4b.MeV)
        beam2.SetParticleMomentumDirection(g4b.G4ThreeVector(par_direction[0],
                                                             par_direction[1],
                                                             par_direction[2]))
        beam2.SetParticleDefinition(electron)
        beam2.SetParticlePosition(g4b.G4ThreeVector(par_in[0]*g4b.um,
                                                    par_in[1]*g4b.um,
                                                    par_in[2]*g4b.um))  
        self.particleGun = beam
        self.particleGun2 = beam2

    def GeneratePrimaries(self, event):
        self.particleGun.GeneratePrimaryVertex(event)
        self.particleGun2.GeneratePrimaryVertex(event)


class MyRunAction(g4b.G4UserRunAction):
    def __init__(self):
        g4b.G4UserRunAction.__init__(self)
        milligray = 1.e-3*g4b.gray
        microgray = 1.e-6*g4b.gray
        nanogray = 1.e-9*g4b.gray
        picogray = 1.e-12*g4b.gray

        g4b.G4UnitDefinition("milligray", "milliGy", "Dose", milligray)
        g4b.G4UnitDefinition("microgray", "microGy", "Dose", microgray)
        g4b.G4UnitDefinition("nanogray", "nanoGy", "Dose", nanogray)
        g4b.G4UnitDefinition("picogray", "picoGy", "Dose", picogray)
      
    def BeginOfRunAction(self, run):
        g4b.G4RunManager.GetRunManager().SetRandomNumberStore(False)
   
    def EndOfRunAction(self, run):
        nofEvents = run.GetNumberOfEvent()
        if nofEvents == 0:
            print("nofEvents=0")
            return

class MyEventAction(g4b.G4UserEventAction):
    "My Event Action"
    def __init__(self, runAction, point_in, point_out):
        g4b.G4UserEventAction.__init__(self)
        self.fRunAction = runAction
        self.point_in = point_in
        self.point_out = point_out

    def BeginOfEventAction(self, event):
        self.edep_device=0.
        self.event_angle = 0.
        self.p_step = []
        self.energy_step = []
        

    def EndOfEventAction(self, event):
        eventID = event.GetEventID()
        print("eventID:%s"%eventID)
        if len(self.p_step):
            point_a = [ b-a for a,b in zip(self.point_in,self.point_out)]
            point_b = [ c-a for a,c in zip(self.point_in,self.p_step[-1])]
            self.event_angle = cal_angle(point_a,point_b)
        else:
            self.event_angle = None
        save_geant4_events(eventID,self.edep_device,
                           self.p_step,self.energy_step,self.event_angle)

    def RecordDevice(self, edep,point_in,point_out):
        self.edep_device += edep
        self.p_step.append([point_in.getX()*1000,
                           point_in.getY()*1000,point_in.getZ()*1000])
        self.energy_step.append(edep)
     

def save_geant4_events(eventID,edep_device,p_step,energy_step,event_angle):
    if(len(p_step)>0):
        s_eventIDs.append(eventID)
        s_edep_devices.append(edep_device)
        s_p_steps.append(p_step)
        s_energy_steps.append(energy_step)
        s_events_angle.append(event_angle)
    else:
        s_eventIDs.append(eventID)
        s_edep_devices.append(edep_device)
        s_p_steps.append([[0,0,0]])
        s_energy_steps.append([0])
        s_events_angle.append(event_angle)

def cal_angle(point_a,point_b):
    "Calculate the anlgle of point a and b"
    x=np.array(point_a)
    y=np.array(point_b)
    l_x=np.sqrt(x.dot(x))
    l_y=np.sqrt(y.dot(y))
    dian=x.dot(y)
    if l_x*l_y > 0:
        cos_=dian/(l_x*l_y)
        angle_d=np.arccos(cos_)*180/np.pi
    else:
        angle_d=9999
    return angle_d


class MySteppingAction(g4b.G4UserSteppingAction):
    "My Stepping Action"
    def __init__(self, eventAction):
        g4b.G4UserSteppingAction.__init__(self)
        self.fEventAction = eventAction

    def UserSteppingAction(self, step):
        edep = step.GetTotalEnergyDeposit()
        point_pre  = step.GetPreStepPoint()
        point_post = step.GetPostStepPoint() 
        point_in   = point_pre.GetPosition()
        point_out  = point_post.GetPosition()
        volume = step.GetPreStepPoint().GetTouchable().GetVolume().GetLogicalVolume()
        volume_name = volume.GetName()
        if(volume_name == "Device"):
            self.fEventAction.RecordDevice(edep,point_in,point_out)


class MyActionInitialization(g4b.G4VUserActionInitialization):
    def __init__(self,par_in,par_out):
        g4b.G4VUserActionInitialization.__init__(self)
        self.par_in = par_in
        self.par_out = par_out

    def Build(self):
        self.SetUserAction(MyPrimaryGeneratorAction(self.par_in,
                                                    self.par_out))
        # global myRA_action
        myRA_action = MyRunAction()
        self.SetUserAction(myRA_action)
        myEA = MyEventAction(myRA_action,self.par_in,self.par_out)
        self.SetUserAction(myEA)
        self.SetUserAction(MySteppingAction(myEA))




# -*- encoding: utf-8 -*-
'''
Description:  Define physical models for different materiales   
@Date       : 2021/09/06 18:46:00
@Author     : yangtao
@version    : 1.0
'''

""" Define Material """

import math

class Material:

    def __init__(self,mat_name):
        self.mat_name = mat_name

    def mat_database(self):

        # material par
        self.si_par_dict = {'Permittivity' : 11.5,\
                             'Avalanche': 'vanOverstraeten',\
                             'Mobility' : 'unknown'\
                            }

        self.sic_par_dict = {'Permittivity' : 9.76,\
                             'Avalanche': 'Hatakeyama',\
                             'Mobility' : 'unknown'\
                            }

        # global data base
        self.mat_db_dict = {'SiC' : self.sic_par_dict,\
                            'Si' : self.si_par_dict\
                            }

        return self.mat_db_dict[self.mat_name]



""" Define Mobility Model """

class Mobility:
    def __init__(self,mat_name):
        self.mat_name = mat_name

    def cal_mobility(self, det, position, charge, electric_field):

        x = position[0]
        y = position[1]
        T = det.temperature # K
        E = electric_field  # V/cm

        doping_expr = det.doping_epr
        doping_expr = doping_expr.replace("x[1]","y")
        doping_expr = doping_expr.replace("sqrt","math.sqrt")
        doping_expr = doping_expr.replace("exp","math.exp")
        #print(doping_expr)
        Neff = abs(eval(doping_expr))

        # SiC mobility
        if(self.mat_name == 'SiC'):
            if(charge>0):
                alpha = 0.34
                ulp = 124 * math.pow(T / 300, -2)
                uminp = 15.9
                Crefp = 1.76e19
                betap = 1.213 * math.pow(T / 300.0, 0.17)
                vsatp = 2e7 * math.pow(T / 300.0, 0.52)
                lfm = uminp + ulp/(1.0 + math.pow(Neff*1e12 / Crefp, alpha))
                hfm = lfm / (math.pow(1.0 + math.pow(lfm * E / vsatp, betap), 1.0 / betap))  

            if(charge<0):
                alpha = 0.61
                ulp = 947 * math.pow(T / 300, -2)
                Crefp = 1.94e19
                betap = 1 * math.pow(T / 300, 0.66)
                vsatp = 2e7 * math.pow(T / 300, 0.87)
                lfm = ulp/ (1 + math.pow(Neff*1e12 / Crefp, alpha))
                hfm = lfm / (math.pow(1.0 + math.pow(lfm * E / vsatp, betap), 1.0/betap))

        # Si mobility
        if(self.mat_name == 'Si'):
            alpha = 0.72*math.pow(T/300.0,0.065)
            if(charge>0):
                ulp = 460.0 * math.pow(T / 300.0, -2.18)
                uminp = 45.0*math.pow(T / 300.0, -0.45)
                Crefp = 2.23e17*math.pow(T / 300.0, 3.2)
                betap = 1.0
                vsatp = 9.05e6 * math.sqrt(math.tanh(312.0/T))
                lfm = uminp + (ulp-uminp)/(1.0 + math.pow(Neff*1e12 / Crefp, alpha))
                hfm = 2*lfm / (1.0+math.pow(1.0 + math.pow(2*lfm * E / vsatp, betap), 1.0 / betap))                        
            else:
                uln = 1430.0 * math.pow(T / 300.0, -2.0)
                uminn = 80.0*math.pow(T / 300.0, -0.45)
                Crefn = 1.12e17*math.pow(T/300.0,3.2)
                betan = 2
                vsatn = 1.45e7 * math.sqrt(math.tanh(155.0/T))
                lfm = uminn + (uln-uminn)/ (1.0 + math.pow(Neff*1e12 / Crefn, alpha))
                hfm = 2*lfm / (1.0+math.pow(1.0 + math.pow(2*lfm * E / vsatn, betan), 1.0/betan))

        return hfm



""" Define Avalanche Model """

class Avalanche:
        
    def __init__(self,model_name):
        self.model_name = model_name

    def cal_coefficient(self, electric_field, charges, temperature):

        coefficient = 0.

        E = electric_field # V/cm
        T = temperature # K

        # van Overstraeten – de Man Model
        if(self.model_name == 'vanOverstraeten'):

            hbarOmega = 0.063 # eV
            E0 = 4.0e5 # V/cm
            T0 = 293.0 # K
            k_T0 = 0.0257 # eV

            # electron
            if( charges < 0 ): 

                a_low = 7.03e5 # cm-1
                a_high = 7.03e5 # cm-1

                b_low = 1.232e6 # cm-1
                b_high = 1.232e6 # cm-1

                #
                # For BandgapDependence parameters
                #

                # Glambda = 62e-8 #cm
                # beta_low = 0.678925 # 1
                # beta_high = 0.678925 # 1

            # hole
            if( charges > 0 ): 

                a_low = 1.582e6 # cm-1
                a_high = 6.71e5 # cm-1

                b_low = 2.036e6 # cm-1
                b_high = 1.693e6 # cm-1

                Glambda = 45e-8 #cm

                beta_low = 0.815009 # 1
                beta_high =  0.677706 # 1

            Ggamma = math.tanh(hbarOmega/(2*k_T0))/math.tanh(hbarOmega/(2*k_T0*T/T0))
            
            if(E>1.75e05):
                if(E>E0):
                    coefficient = Ggamma*a_high*math.exp(-(Ggamma*b_high)/E)
                else:
                    coefficient = Ggamma*a_low*math.exp(-(Ggamma*b_low)/E)
            else:
                coefficient = 0.

    

        if(self.model_name == 'Okuto'):

            T0 = 300.0 # K
            _gamma = 1.0 # 1
            _delta = 2.0 # 1

            # electron
            if( charges < 0):
                a = 0.426 # V-1
                b = 4.81e5 # V/cm
                c = 3.05e-4 # K-1
                d = 6.86e-4 # K-1

                _lambda = 62.0e-8
                _beta = 0.265283

            # hole
            if( charges < 0):
                a = 0.243 # V-1
                b = 6.53e5 # V/cm
                c = 5.35e-4 # K-1
                d = 5.67e-4 # K-1

                _lambda = 45.0e-8 # cm
                _beta = 0.261395 # 1
            
            if(E>1.0e05):
                coefficient = a*(1+c*(T-T0))*pow(E,_gamma)*math.exp(-(b*(1+d*(T-T0)))/E)
            else:
                coefficient = 0.


        if(self.model_name == 'Hatakeyama'):
            '''
            The Hatakeyama avalanche model describes the anisotropic behavior in 4H-SiC power devices. The impact ionization coefficient is obtained according to the Chynoweth law.
            '''
            hbarOmega = 0.19 # eV
            _theta =1 # 1
            T0 = 300.0 # K
            k_T0 = 0.0257 # eV

            if( charges < 0):
                a_0001 = 1.76e8 # cm-1
                a_1120 = 2.10e7 # cm-1
                b_0001 = 3.30e7 # V/cm 
                b_1120 = 1.70e7 # V/cm
                 
            if (charges > 0):
                a_0001 = 3.41e8 # cm-1
                a_1120 = 2.96e7 # cm-1
                b_0001 = 2.50e7 # V/cm 
                b_1120 = 1.60e7 # V/cm 

            _gamma = math.tanh(hbarOmega/(2*k_T0))/math.tanh(hbarOmega/(2*k_T0*T/T0))

            # only consider the <0001> direction multiplication, no anisotropy now!
            a = a_0001
            b = b_0001

            coefficient = _gamma*a*math.exp(-(_gamma*b/E))


        return coefficient


# -*- encoding: utf-8 -*-
'''
Description: 
    Simulate induced current through BB or CSA amplifier 
@Date       : 2021/09/02 14:11:57
@Author     : tanyuhang
@version    : 1.0
'''

import math
import ROOT

# CSA and BB amplifier simulation
class Amplifier:
    def __init__(self,my_d,ampl_par,mintstep="50e-12"):
        """
        Description:
            Get current after CSA and BB amplifer
        Parameters:
        ---------
        CSA_par : dic
            All input paramters of CSA in CSA_par
        BB_par : dic
            All input paramters of BB in CSA_par
        mintsteo : float
            The readout time step (bin width)        
        @Modify:
        ---------
            2021/09/09
        """
        CSA_par = ampl_par[0]
        BB_par = ampl_par[1]
        self.ampli_define(CSA_par,BB_par)
        self.sampling_charge(my_d,mintstep)
        self.ampl_sim() 

    def ampli_define(self,CSA_par,BB_par):
        """
        Description:
            The parameters of CSA and BB amplifier.
            Details introduction can be got in setting module.
        @Modify:
        ---------
            2021/09/09
        """
        self.t_rise    = CSA_par['t_rise']
        self.t_fall    = CSA_par['t_fall']
        self.trans_imp = CSA_par['trans_imp']
        self.CDet      = CSA_par['CDet']
        self.BBW       = BB_par['BBW']
        self.BBGain    = BB_par['BBGain']
        self.BB_imp    = BB_par['BB_imp']
        self.OscBW     = BB_par['OscBW'] 
        ##BB simualtion parameter
        tau_C50 = 1.0e-12*50.*self.CDet          #Oscil. RC
        tau_BW = 0.35/(1.0e9*self.OscBW)/2.2      #Oscil. RC
        tau_BB_RC = 1.0e-12*self.BB_imp*self.CDet     #BB RC
        tau_BB_BW = 0.35/(1.0e9*self.BBW)/2.2    #BB Tau
        self.tau_scope = math.sqrt(pow(tau_C50,2)+pow(tau_BW,2))
        self.tau_BBA =  math.sqrt(pow(tau_BB_RC,2)+pow(tau_BB_BW,2))    #BB_out

    def sampling_charge(self,my_d,mintstep):
        """ Transform current to charge 
        with changing bin width to oscilloscope bin width
        """
        self.max_num = my_d.sum_cu.GetNbinsX()
        self.max_hist_num = my_d.n_bin
        self.undersampling = int(float(mintstep)/my_d.t_bin)
        self.time_unit = my_d.t_bin*self.undersampling
        self.CDet_j = 0     # CSA readout mode
        self.itot = [0.0]*self.max_num
        self.qtot = 0.0
        # get total charge
        i=0
        for j in range(0,self.max_hist_num,self.undersampling):
            self.itot[i] = my_d.sum_cu.GetBinContent(j)
            self.qtot = self.qtot + self.itot[i]*self.time_unit
            i+=1
        max_hist_num = int(self.max_hist_num/self.undersampling)
        IintTime = max(2.0*(self.t_rise+self.t_fall)*1e-9/self.time_unit,
                       3.0*self.tau_BBA/self.time_unit)
        self.IMaxSh = int(max_hist_num + IintTime)

    def ampl_sim(self):
        """
        Description:
            CSA and BB amplifier Simulation         
        Parameters:
        ---------
        arg1 : int
            
        @Modify:
        ---------
            2021/09/09
        """
        IMaxSh = self.IMaxSh
        preamp_Q = [0.0]*IMaxSh
        self.CSA_p_init()
        self.BB_p_init()
        step=1
        for i in range(IMaxSh-step):
            if(i>0 and i <self.max_hist_num-step):
                preamp_Q[i] = 0.0
                for il in range(i,i+step):
                    preamp_Q[i] += self.itot[il]*self.time_unit
            elif (i != 0):
                preamp_Q[i]=0.0

        for i in range(IMaxSh-step):
            if i >= step:
                dif_shaper_Q = preamp_Q[i]
            else:
                dif_shaper_Q = 0
            for j in range(IMaxSh-i):
                self.fill_CSA_out(i,j,dif_shaper_Q)
                self.fill_BB_out(i,j,dif_shaper_Q)
            self.max_CSA(i)
        self.fill_CSA_th1f()
        self.fill_BB_th1f()

    def CSA_p_init(self):
        """ CSA parameter initialization"""
        t_rise = self.t_rise
        t_fall = self.t_fall
        self.tau_rise = t_rise/2.2*1e-9
        self.tau_fall = t_fall/2.2*1e-9
        if (self.tau_rise == self.tau_fall):
            self.tau_rise *= 0.9
        self.sh_max = 0.0  
        self.shaper_out_Q = [0.0]*self.IMaxSh
        self.shaper_out_V = [0.0]*self.IMaxSh

    def BB_p_init(self):
        """ BB parameter initialization"""
        self.Vout_scope = [0.0]*self.IMaxSh
        self.Iout_BB_RC = [0.0]*self.IMaxSh
        self.Iout_C50 = [0.0]*self.IMaxSh   
        self.BBGraph = [0.0]*self.IMaxSh

    def fill_CSA_out(self,i,j,dif_shaper_Q):
        """ Fill CSA out variable"""     
        self.shaper_out_Q[i+j] += self.tau_fall/(self.tau_fall+self.tau_rise) \
                                  * dif_shaper_Q*(math.exp(-j*self.time_unit
                                  / self.tau_fall)-math.exp(
                                  - j*self.time_unit/self.tau_rise))

    def fill_BB_out(self,i,j,dif_shaper_Q):
        """ Fill BB out variable"""   
        self.Iout_C50[i+j] += (dif_shaper_Q)/self.tau_scope \
                                * math.exp(-j*self.time_unit/self.tau_scope)
        self.Iout_BB_RC[i+j] += (dif_shaper_Q)/self.tau_BBA \
                                * math.exp(-j*self.time_unit/self.tau_BBA)
        self.BBGraph[i+j] = 1e+3 * self.BBGain * self.Iout_BB_RC[i+j]
        self.Vout_scope[i+j] = 50 * self.Iout_C50[i+j]
        if (abs(self.BBGraph[i+j]) > 800):
            self.BBGraph[i+j] = 800*self.BBGraph[i+j]/abs(self.BBGraph[i+j])

    def max_CSA(self,i):
        """ Get max out value of CSA"""               
        if (abs(self.shaper_out_Q[i]) > abs(self.sh_max)):
            self.sh_max = self.shaper_out_Q[i]

    def fill_CSA_th1f(self):
        """ Change charge to amplitude [mV]
            and save in the th1f
        """
        Ci = 3.5e-11  #fF
        Qfrac = 1.0/(1.0+self.CDet*1e-12/Ci)
        self.CSA_ele = ROOT.TH1F("electronics", "electronics",
                                 self.IMaxSh, 0, self.IMaxSh*self.time_unit)
        for i in range(self.IMaxSh):
            if self.sh_max == 0.0:
                self.shaper_out_V[i] = 0.0
            elif self.CDet_j == 0:
                self.shaper_out_V[i] = self.shaper_out_Q[i]*self.trans_imp\
                                       * 1e15*self.qtot*Qfrac/self.sh_max            
            elif self.CDet_j ==1:
                self.shaper_out_V[i] = self.shaper_out_Q[i]*self.trans_imp\
                                       * 1e15*self.qtot/self.sh_max
            self.CSA_ele.SetBinContent(i,self.shaper_out_V[i])
        #Print the max current time of CSA
        max_CSA_height = min(self.shaper_out_V)
        time_t = self.shaper_out_V.index(max_CSA_height)
        print("CSA_time=%s" %(time_t*self.time_unit))

    def fill_BB_th1f(self):
        """ Change charge to amplitude [mV]
            and save in the th1f
        """
        self.BB_ele = ROOT.TH1F("electronics BB","electronics BB",
                                self.IMaxSh,0,self.IMaxSh*self.time_unit)
        for i in range(len(self.BBGraph)+1):
            if i == 0:
                self.BB_ele.SetBinContent(i,0)
            else:
                self.BB_ele.SetBinContent(i,self.BBGraph[i-1])
        # Print the max current time of BB
        max_BB_height = min(self.BBGraph)
        time_t = self.BBGraph.index(max_BB_height)
        print("BB_time=%s"%(time_t*self.time_unit))
          
    def __del__(self):
        pass


# -*- encoding: utf-8 -*-
'''
Description:  PyPI     
@created   : 2021/09/08 09:33:59
'''

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="raser",
    version="2.2.1",
    author="Xin Shi",
    author_email="Xin.Shi@outlook.com",
    description="SiC Detector Simulation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dt-np/raser",
    project_urls={
        "Bug Tracker": "https://github.com/dt-np/raser/issues",
        "Documentation": "http://xteam.ihep.ac.cn/research/rd50/raser/",
    }, 
    packages=find_packages(),
    license='MIT',
    classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
    ]
	)




from sys import modules
from raser.geometry import R3dDetector
from raser.pyfenics import FenicsCal
from raser.g4particles import Particles
from raser.calcurrent import CalCurrent
from raser.elereadout import Amplifier
from raser.setting import Setting

from raser.geometry import R2dDetector
from raser.pyfenics import FenicsCal2D
from raser.calcurrent import CalCurrent2D
from raser.model import Material
from raser.model import Avalanche
from raser.source import MIPs
